# coding: utf-8
from core.logging.logger_utils import log
from core.model.registered import Registered
from core.utils.masking_message import masking

import scenarios.logging.logger_constants as log_const

field_models = Registered()


def field_model_factory(description, raw_data, user, lifetime):
    _type = type(description)
    model = field_models[_type]
    instance = model(description, raw_data, user, lifetime)
    return instance


class BasicField:

    def __init__(self, description, items, user, lifetime):
        items = items or {}
        self.description = description
        self._value = items.get("value")
        self._available = items.get("available", self.description.available)
        self._user = user
        self._lifetime = lifetime
        self._masking_fields = user.settings["template_settings"].get("masking_fields") if user is not None else []

    @property
    def value(self):
        return self._value

    @property
    def available(self):
        return self._available

    @property
    def can_be_updated(self):
        return self.value is not None

    def check_can_be_filled(self, text_preprocessing_result, user):
        return (
                self.description.requirement.check(text_preprocessing_result, user) and
                self.description.filler.run(user, text_preprocessing_result) is not None
        )

    @property
    def valid(self):
        return self.value is not None or not self.description.required

    def is_fill_need(self, value, origin_value):
        return value is not None

    def fill(self, origin_value):
        filled = False
        value = origin_value if origin_value is not None else self.description.default_value
        if self.is_fill_need(value, origin_value):
            self._set_value(value)
            self.reset_available()
            filled = True
        return filled

    def _set_value(self, value):
        self._value = value
        dict_value = {self.description.name: value}
        masked_dict_value = masking(dict_value, self._masking_fields)
        message = "%(class_name)s: %(description_id)s filled by value: %(field_value)s"
        params = {log_const.KEY_NAME: log_const.FILLER_RESULT_VALUE,
                  "class_name": self.__class__.__name__,
                  "description_id": self.description.id,
                  "field_value": str(masked_dict_value[self.description.name])}
        log(message, None, params)

    def set_available(self):
        self._available = True

    def reset_available(self):
        self._available = self.description.available

    @property
    def raw(self):
        result = {}
        is_value_changed = (self._value is not None and self._value != self.description.default_value)
        if is_value_changed:
            result["value"] = self._value
        if self._available != self.description.available:
            result["available"] = self._available
        return result


class QuestionField(BasicField):
    def __init__(self, description, items, user, lifetime):
        items = items or {}
        super(QuestionField, self).__init__(description, items, user, lifetime)
        self.ask_again_counter = items.get("ask_again_counter", 0)

    @property
    def value(self):
        return self._value if self._value is not None else self.default_value

    @property
    def default_value(self):
        if self.description.need_load_context:
            prev_value = self._user.last_fields[self.description.id].value
            if prev_value is not None:
                return prev_value
        return self.description.default_value

    def is_fill_need(self, value, origin_value):
        check_new_field_value = self._value != value and origin_value is not None
        can_set = value is not None and (self._value is None or check_new_field_value)
        return self.available and can_set

    def _set_value(self, value):
        super()._set_value(value)
        if self.description.need_save_context:
            self._user.last_fields[self.description.id].value = value
            self._user.last_fields[self.description.id].set_remove_time(self._lifetime)

    @property
    def raw(self):
        result = super().raw
        if self.ask_again_counter != 0:
            result["ask_again_counter"] = self.ask_again_counter
        return result


class IntegrationField(BasicField):
    def __init__(self, description, items, user, lifetime):
        items = items or {}
        super(IntegrationField, self).__init__(description, items, user, lifetime)

    # TODO сделать required requirement
    # def valid(self, text_preprocessing_result, user, params):
    #     return self.value is not None or not self.description.required.check(text_preprocessing_result, user, params)



from unittest import TestCase
from unittest.mock import Mock

from core.utils.rerunable import Rerunable
from smart_kit.utils.picklable_mock import PicklableMock


class HandledException(Exception):
    pass


class AllFailedException(Exception):
    pass


class CustomException(Exception):
    pass


class RerunableOne(Rerunable):
    def __init__(self, try_count):
        super(RerunableOne, self).__init__({"try_count": try_count})
        self._on_prepare_mock = PicklableMock()

    @property
    def _handled_exception(self):
        return HandledException

    def _on_prepare(self):
        self._on_prepare_mock()

    def _on_all_tries_fail(self):
        raise AllFailedException

    def run(self, *args, **kwargs):
        return self._run(*args, **kwargs)


class TestRerunable(TestCase):
    def setUp(self):
        self.try_count = 2
        self.rerunable = RerunableOne(self.try_count)
        self.expected_value = PicklableMock()
        self.param = PicklableMock()
        self.param1 = {"param1_name": PicklableMock()}

    def test_pass(self):
        self.action = PicklableMock()
        self.rerunable.run(self.action, self.param, **self.param1)
        self.action.assert_called_once_with(self.param, **self.param1)

    def test_wrong_exception(self):
        self.action = Mock(side_effect=CustomException())
        with self.assertRaises(CustomException) as context:
            self.rerunable.run(self.action, self.param, **self.param1)
        self.action.assert_called_once_with(self.param, **self.param1)

    def test_all_retry_failed(self):
        self.action = Mock(side_effect=HandledException())
        with self.assertRaises(AllFailedException) as context:
            self.rerunable.run(self.action, self.param, **self.param1)
        self.action.assert_called_with(self.param, **self.param1)
        self.assertEqual(self.action.call_count, self.try_count)

    def test_first_try_failed(self):
        self.param = PicklableMock()
        self.param1 = {"param1_name": PicklableMock()}
        self.action = Mock(side_effect=[HandledException(), self.expected_value])
        result = self.rerunable.run(self.action, self.param, **self.param1)
        self.action.assert_called_with(self.param, **self.param1)
        self.assertEqual(self.action.call_count, self.try_count)
        self.assertEqual(result, self.expected_value)

    def test_first_try_ok(self):
        self.action = Mock(return_value=self.expected_value)
        result = self.rerunable.run(self.action, self.param, **self.param1)
        self.action.assert_called_once_with(self.param, **self.param1)
        self.assertEqual(result, self.expected_value)



from typing import Dict, Any, Optional, List, Union

from core.basic_models.actions.basic_actions import Action
from core.basic_models.actions.string_actions import StringAction
from core.basic_models.actions.command import Command
from core.text_preprocessing.base import BaseTextPreprocessingResult

from scenarios.user.user_model import User

from smart_kit.names.message_names import NOTHING_FOUND


class NothingFoundAction(Action):
    version: Optional[int]
    id: Optional[str]

    def __init__(self, items: Dict[str, Any] = None, id: Optional[str] = None):
        super().__init__(items, id)
        self._action = StringAction({"command": NOTHING_FOUND})

    def run(self, user: User, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        commands.extend(self._action.run(user, text_preprocessing_result, params=params))
        return commands



import json
from unittest import TestCase

from core.message.from_message import SmartAppFromMessage
from core.message.msg_validator import MessageValidator
from core.utils.utils import current_time_ms


class PieMessageValidator(MessageValidator):
    def validate(self, message_name: str, payload: dict):
        return 3.14 < payload.get("pi", 0) < 3.15


class TestFromMessage(TestCase):
    def test_1(self):
        input_msg = {
            "messageId": 2,
            "uuid": {"userChannel": "B2C", "userId": "userId", "sub": "sub"},
            "payload": {
                "message": {
                    "original_text": "сверни приложение"
                },
                "device": {
                    "platformType": "android",
                    "platformVersion": "9",
                    "surface": "STARGATE",
                    "surfaceVersion": "1.56.20200828144304",
                    "features": {"appTypes": ["APK", "WEB_APP", "DIALOG"]},
                    "capabilities": {"mic": {"available": True},
                                     "screen": {"available": True},
                                     "speak": {"available": True}},
                    "deviceId": "34534545345345",
                    "deviceManufacturer": "SberDevices",
                    "deviceModel": "stargate",
                    "additionalInfo": {}
                }
            },
            "messageName": "MESSAGE_TO_SKILL"
        }
        topic = "test"
        headers = []
        current_time = current_time_ms()
        message = SmartAppFromMessage(value=input_msg, topic_key=topic, headers=headers, creation_time=current_time)

        self.assertAlmostEqual(message.creation_time, current_time)
        self.assertEqual(2, message.incremental_id)
        self.assertEqual(input_msg["uuid"]["userChannel"], message.channel)
        self.assertEqual(input_msg["messageName"], message.type)
        self.assertEqual(input_msg["uuid"]["userId"], message.uid)
        self.assertEqual(json.dumps(input_msg, ensure_ascii=False), message.as_str)
        self.assertEqual("userId_B2C", message.db_uid)
        self.assertDictEqual(input_msg["uuid"], message.uuid)
        self.assertDictEqual(input_msg["payload"], message.payload)
        device = input_msg["payload"]["device"]
        self.assertEqual(device["platformType"], message.device.platform_type)
        self.assertEqual(device["platformVersion"], message.device.platform_version)
        self.assertEqual(device["surface"], message.device.surface)
        self.assertEqual(device["surfaceVersion"], message.device.surface_version)
        self.assertEqual(device["features"], message.device.features)
        self.assertEqual(device["capabilities"], message.device.capabilities)
        self.assertEqual(device["additionalInfo"], message.device.additional_info)
        self.assertEqual(topic, message.topic_key)

    def test_valid_true(self):
        input_msg = {
            "messageId": 2,
            "sessionId": "234",
            "uuid": {"userChannel": "web", "userId": "99", "chatId": "80"},
            "payload": {"key": "some payload"},
            "messageName": "some_type"
        }
        headers = [('test_header', 'result')]
        message = SmartAppFromMessage(input_msg, headers=headers)
        self.assertTrue(message.validate())

    def test_valid_false(self):
        input_msg = {
            "uuid": {"userChannel": "web", "userId": 99, "chatId": 80},
            "payload": "some payload"
        }
        headers = [('test_header', 'result')]

        message = SmartAppFromMessage(input_msg, headers=headers)
        self.assertFalse(message.validate())

    def test_validation(self):
        input_msg = {
            "uuid": {"userChannel": "web", "userId": "99", "chatId": "80"},
            "messageName": "some_name",
            "messageId": "random_id",
            "sessionId": "random_id",
            "payload": {
                "pi": 3.14159265358979,
            }
        }
        headers = [('test_header', 'result')]

        message = SmartAppFromMessage(
            input_msg,
            headers=headers, validators=(PieMessageValidator(),))
        self.assertTrue(message.validate())

        input_msg["payload"]["pi"] = 2.7182818284
        message = SmartAppFromMessage(
            input_msg,
            headers=headers, validators=(PieMessageValidator(),))
        self.assertFalse(message.validate())



import inspect
import logging
import re
from unittest.mock import Mock
from typing import Dict, List, Union, Optional

import timeout_decorator

import core.basic_models.classifiers.classifiers_constants as cls_const
import core.logging.logger_constants as log_const
import scenarios.logging.logger_constants as scenarios_log_const
from core.basic_models.classifiers.basic_classifiers import Classifier
from core.utils.masking_message import masking
from core.utils.stats_timer import StatsTimer

MESSAGE_ID_STR = "message_id"
UID_STR = "uid"
LOGGING_UUID = "logging_uuid"
CLASS_NAME = "class_name"
LOG_STORE_FOR = "log_store_for"
HEADERS = "headers"
CALLER_INFO = "caller_info"


class LoggerMessageCreator:
    LOGGER_HEADERS = ["kafka_replyTopic", "app_callback_id"]
    ART_NAMES = [
        "channel", "type", "device_channel", "device_channel_version", "device_platform", "group",
        "device_platform_version", "device_platform_client_type", "csa_profile_id", "test_deploy"
    ]

    @classmethod
    def update_user_params(cls, user, params):
        message = user.message
        for name in cls.ART_NAMES:
            param = getattr(message, name, None)
            if param:
                params[name] = param

    @classmethod
    def update_other_params(cls, user, params, cls_name='', log_store_for=1):
        message_id, uuid, logging_uuid = None, None, None
        if user:
            message = user.message
            message_id = message.incremental_id
            uuid = user.id
            logging_uuid = message.logging_uuid
        params.setdefault(UID_STR, uuid)
        params.setdefault(MESSAGE_ID_STR, message_id)
        params.setdefault(LOGGING_UUID, logging_uuid)
        params[CLASS_NAME] = cls_name
        params[LOG_STORE_FOR] = log_store_for

    @classmethod
    def escape(cls, string):
        return re.sub(r"(%[^\(])", r"%\1", string)

    @classmethod
    def make_message(cls, user=None, params=None, cls_name='', log_store_for=1):
        params = params or {}
        cls.filter_headers(params)
        if HEADERS in params:
            params[HEADERS] = str(params[HEADERS])
        if user:
            cls.update_user_params(user, params)
        masked_params = masking(params)
        cls.update_other_params(user, masked_params, cls_name, log_store_for)
        return masked_params

    @classmethod
    def filter_headers(cls, params: dict):
        if HEADERS in params:
            if isinstance(params[HEADERS], list):
                params[HEADERS] = [(key, value) for key, value in params[HEADERS] if key in cls.LOGGER_HEADERS]


default_logger = logging.getLogger()


def log(message, user=None, params=None, level="INFO", exc_info=None, log_store_for=1):
    try:
        level_name = logging.getLevelName(level)
        current_frame = inspect.currentframe()
        previous_frame = current_frame.f_back
        module_name = previous_frame.f_globals["__name__"]
        logger = logging.getLogger(module_name)
        if not logger.isEnabledFor(level_name):
            return

        caller = inspect.getframeinfo(previous_frame)
        caller_info = {"filename": caller.filename, "function": caller.function, "line_num": caller.lineno}
        params = params or {}
        params[CALLER_INFO] = caller_info

        instance = previous_frame.f_locals.get('self', None)

        from smart_kit.configs import get_app_config
        try:
            message_maker = get_app_config().LOGGER_MESSAGE_CREATOR
            if isinstance(message_maker, Mock):
                raise AttributeError
        except AttributeError:
            message_maker = LoggerMessageCreator

        # cast log_params["params"] to str
        # TODO: think how to make it more general
        if params and "params" in params:
            params["params"] = str(params["params"])

        log_store_for_map = getattr(logging, "log_store_for_map", None)
        if log_store_for_map is not None and params is not None:
            log_store_for = log_store_for_map.get(params.get(log_const.KEY_NAME), log_store_for)

        if instance is not None:
            params = message_maker.make_message(user, params, instance.__class__.__name__, log_store_for)
        else:
            params = message_maker.make_message(user, params, log_store_for=log_store_for)

        # эскейпим сишное форматирование логгера,
        # см. tests.core_tests.test_utils.test_logger.TestLogger.test_escaping
        message = message_maker.escape(message)

        logger.log(level_name, message, params, exc_info=exc_info)
    except timeout_decorator.TimeoutError:
        raise
    except:
        default_logger.log(logging.getLevelName("ERROR"), "Failed to write a log. Exception occurred",
                           params, exc_info=True, stack_info=True)


def log_classifier_result(classification_res: List[Dict[str, Union[str, float, bool]]], user,
                          classifier: Classifier, timer: Optional[StatsTimer] = None) -> None:
    classifier_name = classifier.settings.get("classifier", "intent_recognizer")
    score_key = cls_const.INTENT_RECOGNIZER_ANSWER_DISTANCE_KEY if classifier_name == "intent_recognizer" \
        else cls_const.SCORE_KEY
    params = {
        log_const.KEY_NAME: scenarios_log_const.CLASSIFIER_VALUE,
        "classifier_name": classifier_name,
        "result": [el[cls_const.ANSWER_KEY] for el in classification_res],
        "score": [el[score_key] for el in classification_res]
    }
    if timer:
        params["time"] = timer.msecs

    log(scenarios_log_const.CLASSIFIER_MESSAGE, user, params)


# backward naming_compatibility
behaviour_log = log



import unittest
from unittest.mock import Mock

from core.basic_models.parametrizers.parametrizer import BasicParametrizer
from smart_kit.utils.picklable_mock import PicklableMock


class ParametrizerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.user = Mock(message=PicklableMock())

    def test_get_user_data(self):
        expected = ["message"]
        parametrizer = BasicParametrizer(self.user, {})
        result = parametrizer._get_user_data(None)
        self.assertListEqual(expected, list(result.keys()))

    def test_collect(self):
        expected = ["message"]
        parametrizer = BasicParametrizer(self.user, {})
        result = parametrizer._get_user_data(None)
        self.assertListEqual(expected, list(result.keys()))



# coding=utf-8
from collections import OrderedDict

import json


def ordered_json(data):
    return json.loads(data, object_pairs_hook=OrderedDict)


def reverse_json_dict(data):
    data = json.loads(data)
    result = dict()
    for key, values in data.items():
        for value in values:
            result[value] = key
    return result



from unittest import TestCase
from unittest.mock import Mock

from scenarios.scenario_models.field.field_filler_description import ExternalFieldFillerDescription
from smart_kit.utils.picklable_mock import PicklableMock


class TestExternalFieldFillerDescription(TestCase):

    def test_1(self):
        expected = 5
        items = {"filler": "my_key"}
        mock_filler = PicklableMock()
        mock_filler.run = Mock(return_value=expected)

        mock_user = PicklableMock()
        mock_user.descriptions = {"external_field_fillers": {"my_key": mock_filler}}

        filler = ExternalFieldFillerDescription(items)
        result = filler.extract(None, mock_user)

        self.assertEqual(expected, result)



# coding: utf-8

from core.model.factory import ordered_dict_factory, build_factory
from core.model.registered import Registered
from scenarios.scenario_models.field.field_descriptions.basic_field_description import BasicFieldDescription

form_descriptions = Registered()

form_description_factory = build_factory(form_descriptions)


class BaseFormDescription:
    DEFAULT_FORM_LIFETIME = 1 * 24 * 60 * 60

    def __init__(self, items, id):
        self.id = id
        self.valid = items.get("valid", False)
        self.lifetime = items.get("lifetime", self.DEFAULT_FORM_LIFETIME)
        self.version = items.get("version", -1)
        self.form_description = items.get("form_description", "")


class FormDescription(BaseFormDescription):
    def __init__(self, items, id):
        super(FormDescription, self).__init__(items, id)
        self._fields = items["fields"]
        self.fields = self.build_fields()

    @ordered_dict_factory(BasicFieldDescription)
    def build_fields(self):
        return self._fields


class CompositeFormDescription(BaseFormDescription):
    def __init__(self, items, id):
        super(CompositeFormDescription, self).__init__(items, id)
        self._forms = items["forms"]
        self.forms = self.build_forms()

    @ordered_dict_factory(BaseFormDescription)
    def build_forms(self):
        return self._forms



# coding: utf-8
import re

from core.basic_models.requirement.basic_requirements import Requirement
from core.text_preprocessing.base import BaseTextPreprocessingResult
from core.unified_template.unified_template import UnifiedTemplate
from scenarios.user.user_model import User
from typing import Optional, Dict, Any


class AskAgainExistRequirement(Requirement):

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
              params: Dict[str, Any] = None) -> bool:
        last_scenario_id = user.last_scenarios.last_scenario_name
        scenario = user.descriptions["scenarios"].get(last_scenario_id)
        return scenario.check_ask_again_requests(text_preprocessing_result, user, params)


class TemplateInArrayRequirement(Requirement):
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(TemplateInArrayRequirement, self).__init__(items, id)
        self._template = UnifiedTemplate(items["template"])
        self._items = set(items["items"])

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
              params: Dict[str, Any] = None) -> bool:
        params = params or {}
        collected = user.parametrizer.collect(text_preprocessing_result)
        params.update(collected)
        render_result = self._template.render(params)
        return render_result in self._items


class ArrayItemInTemplateRequirement(Requirement):
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(ArrayItemInTemplateRequirement, self).__init__(items, id)
        self._template = UnifiedTemplate(items["template"])
        self._items = set(items["items"])

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
              params: Dict[str, Any] = None) -> bool:
        params = params or {}
        collected = user.parametrizer.collect(text_preprocessing_result)
        params.update(collected)
        render_result = self._template.render(params)
        for item in self._items:
            if item in render_result:
                return True
        return False


class RegexpInTemplateRequirement(Requirement):
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(RegexpInTemplateRequirement, self).__init__(items, id)
        self._template = UnifiedTemplate(items["template"])
        self._regexp = re.compile(items["regexp"], re.S | re.M)

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
              params: Dict[str, Any] = None) -> bool:
        params = params or {}
        collected = user.parametrizer.collect(text_preprocessing_result)
        params.update(collected)
        render_result = self._template.render(params)
        return True if self._regexp.search(render_result) else False



# coding: utf-8
import unittest
from unittest.mock import Mock

import scenarios.logging.logger_constants as log_const
from scenarios.actions.action import RunScenarioByProjectNameAction


class TestScenarioDesc(dict):
    def run(self, argv1, argv2, params):
        return ['result to run scenario']


class RunScenarioByProjectNameActionTest1(unittest.TestCase):
    def setUp(self):
        self.test_text_preprocessing_result = Mock('text_preprocessing_result')
        self.test_user1 = Mock('User')
        self.test_user1.name = "TestName"
        self.test_user1.message = Mock('message')
        self.test_user1.message.project_name = 'any scenario'
        self.test_user1.message.channel = 'test_channel'
        self.test_user1.descriptions = {"scenarios": {"any scenario": TestScenarioDesc([[1, 1]])}}
        self.test_user2 = Mock('User')
        self.test_user2.name = "TestName"
        self.test_user2.id = 222
        self.test_user2.message = Mock('message')
        self.test_user2.message.incremental_id = 111
        self.test_user2.message.logging_uuid = 111
        self.test_user2.message.project_name = 'specific scenario'
        self.test_user2.message.channel = 'test_channel'
        self.test_user2.descriptions = {"scenarios": {"any scenario": TestScenarioDesc()}}
        self.test_text_preprocessing_result = Mock('TextPreprocessingResult')
        self.items = {"any_key": "any value"}

    def test_run_scenario_by_project_name_run(self):
        obj1 = RunScenarioByProjectNameAction(self.items)
        # без оглядки на аннотации из PEP 484
        self.assertEqual(obj1.run(self.test_user1, self.test_text_preprocessing_result, {'any_attr': {'any_data'}}),
                        ['result to run scenario'])
        obj2 = RunScenarioByProjectNameAction(self.items)
        self.assertEqual([], obj2.run(self.test_user2, self.test_text_preprocessing_result))

    def test_run_scenario_by_project_name_log_vars(self):
        obj = RunScenarioByProjectNameAction(self.items)
        self.assertIsNotNone(log_const.KEY_NAME)
        self.assertIsNotNone(obj.__class__.__name__)



import json
from typing import Optional, Dict, Union, List, Any

import requests

import core.logging.logger_constants as log_const
from core.basic_models.actions.command import Command
from core.basic_models.actions.string_actions import NodeAction
from core.logging.logger_utils import log
from core.model.base_user import BaseUser
from core.text_preprocessing.base import BaseTextPreprocessingResult


class HTTPRequestAction(NodeAction):
    """
    Example:
        {
            "params": {
                "method": "POST",
                "url": "https://some_url.com/...",
                ... (см. BaseHttpRequestAction)
            },
            "store": "..."  // название переменной в user.variables, куда сохранится результат
            "behavior": "..."  // название behavior'a, вызываемого после исполнения запроса
        }
    """

    POST = "POST"
    GET = "GET"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"
    DEFAULT_METHOD = POST
    DEFAULT_TIMEOUT = 3

    TIMEOUT = "TIMEOUT"
    CONNECTION = "CONNECTION"

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super().__init__(items, id)
        self.method_params = items['params']
        self.method_params.setdefault("method", self.DEFAULT_METHOD)
        self.error = None
        self.init_save_params(items)

    def init_save_params(self, items):
        self.store = items["store"]
        self.behavior = items.get("behavior")

    def preprocess(self, user, text_processing, params):
        behavior_description = user.descriptions["behaviors"].get(self.behavior)
        timeout = behavior_description.timeout(user) if behavior_description else self.DEFAULT_TIMEOUT
        self.method_params.setdefault("timeout", timeout)

    @staticmethod
    def _check_headers_validity(headers: Dict[str, Any], user) -> Dict[str, str]:
        for header_name, header_value in list(headers.items()):
            if not isinstance(header_value, (str, bytes)):
                if isinstance(header_value, (int, float, bool)):
                    headers[header_name] = str(header_value)
                else:
                    log(f"{__class__.__name__}._check_headers_validity remove header {header_name} because "
                        f"({type(header_value)}) is not in [int, float, bool, str, bytes]", user=user, params={
                        log_const.KEY_NAME: "sent_http_remove_header",
                    })
                    del headers[header_name]
        return headers

    def _make_response(self, request_parameters, user):
        try:
            with requests.request(**request_parameters) as response:
                response.raise_for_status()
                self._log_response(user, response)
                return response
        except requests.exceptions.Timeout:
            self.error = self.TIMEOUT
        except (requests.exceptions.ConnectionError, requests.exceptions.HTTPError):
            self.error = self.CONNECTION

    def _get_request_params(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
                            params: Optional[Dict[str, Union[str, float, int]]] = None):
        collected = user.parametrizer.collect(text_preprocessing_result)
        params.update(collected)
        request_parameters = self._get_rendered_tree_recursive(self._get_template_tree(self.method_params), params)
        req_headers = request_parameters.get("headers")
        if req_headers:
            # Заголовки в запросах должны иметь тип str или bytes. Поэтому добавлена проверка и приведение к типу str,
            # на тот случай если в сценарии заголовок указали как int, float и тд
            request_parameters["headers"] = self._check_headers_validity(req_headers, user)
        return request_parameters

    def _log_request(self, user, request_parameters, additional_params=None):
        additional_params = additional_params or {}
        log(f"{self.__class__.__name__}.run sent https request ", user=user, params={
            **request_parameters,
            log_const.KEY_NAME: "sent_http_request",
            **additional_params,
        })

    def _log_response(self, user, response, additional_params=None):
        additional_params = additional_params or {}
        log(f"{self.__class__.__name__}.run get https response ", user=user, params={
            'headers': dict(response.headers),
            'time': response.elapsed.microseconds,
            'cookie': {i.name: i.value for i in response.cookies},
            'status': response.status_code,
            log_const.KEY_NAME: "got_http_response",
            **additional_params,
        })

    def process_result(self, response, user, text_preprocessing_result, params):
        behavior_description = user.descriptions["behaviors"][self.behavior] if self.behavior else None
        action = None
        if self.error is None:
            try:
                data = response.json()
            except json.decoder.JSONDecodeError:
                data = None
            user.variables.set(self.store, data)
            action = behavior_description.success_action if behavior_description else None
        elif behavior_description:
            if self.error == self.TIMEOUT:
                action = behavior_description.timeout_action
            else:
                action = behavior_description.fail_action
        return action.run(user, text_preprocessing_result, None) if action else None

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> Optional[List[Command]]:
        self.preprocess(user, text_preprocessing_result, params)
        params = params or {}
        request_parameters = self._get_request_params(user, text_preprocessing_result, params)
        self._log_request(user, request_parameters)
        response = self._make_response(request_parameters, user)
        return self.process_result(response, user, text_preprocessing_result, params)



# coding: utf-8
from typing import Dict, Any

from core.basic_models.scenarios.base_scenario import BaseScenario
from core.monitoring.monitoring import monitoring
from core.logging.logger_utils import log

import scenarios.logging.logger_constants as log_const
from scenarios.scenario_models.field.field import QuestionField
from scenarios.scenario_models.history import Event, HistoryConstants
from scenarios.actions.action_params_names import REQUEST_FIELD

FORM_FIELD_DELIMETER = "__"


class FormFillingScenario(BaseScenario):
    def __init__(self, items, id):
        super(FormFillingScenario, self).__init__(items, id)
        self.form_type = items["form"]
        self.keep_forms_alive = items.get("keep_form_alive", False)
        self.tag = items.get("tag")

    def _get_form(self, user):
        form = user.forms.get_or_create(self.form_type)
        form = user.forms.new(self.form_type) if form.is_valid() else form
        form.refresh()
        return form

    def text_fits(self, text_preprocessing_result, user):
        return self._check_field(text_preprocessing_result, user, None)

    def check_ask_again_requests(self, text_preprocessing_result, user, params):
        form = user.forms[self.form_type]
        question_field = self._field(form, text_preprocessing_result, user, params)
        return question_field.ask_again_counter < len(question_field.description.ask_again_requests)

    def ask_again(self, text_preprocessing_result, user, params):
        form = user.forms[self.form_type]
        question_field = self._field(form, text_preprocessing_result, user, params)
        question = question_field.description.ask_again_requests[question_field.ask_again_counter]
        question_field.ask_again_counter += 1

        user.history.add_event(
            Event(type=HistoryConstants.types.FIELD_EVENT,
                  scenario=self.root_id,
                  content={HistoryConstants.content_fields.FIELD: question_field.description.id},
                  result=HistoryConstants.event_results.ASK_QUESTION))

        return question.run(user, text_preprocessing_result, params)

    def _check_field(self, text_preprocessing_result, user, params):
        form = user.forms[self.form_type]
        field = self._field(form, text_preprocessing_result, user, params)
        return field.check_can_be_filled(text_preprocessing_result, user) if field else False

    def _field(self, form, text_preprocessing_result, user, params):
        return self._find_field(form, text_preprocessing_result, user, params)

    def _find_field(self, form, text_preprocessing_result, user, params):
        for field_name in form.fields.descriptions:
            field = form.fields[field_name]
            if not field.valid and field.description.has_requests and \
                    field.description.requirement.check(
                    text_preprocessing_result, user, params
                    ):
                return field

    def get_fields_data(self, form, form_key):
        form_fields_data = {}
        for field_key, field_value in form.fields.values.items():
            key = "".join((self._clean_key(form_key), FORM_FIELD_DELIMETER, self._clean_key(field_key)))
            form_fields_data[key] = field_value
        return form_fields_data

    def _clean_key(self, key: str):
        return key.replace(" ", "")

    def _extract_by_field_filler(self, field_key, field_descr, text_normalization_result, user, params):
        result = {}
        check = field_descr.requirement.check(text_normalization_result, user, params)
        log_params = self._log_params()
        log_params["requirement"] = field_descr.requirement.__class__.__name__,
        log_params["check"] = check
        log_params["field_key"] = field_key
        message = "FormFillingScenario.extract: field %(field_key)s requirement %(requirement)s return value: %(check)s"
        log(message, user, log_params)
        if check:
            result[field_key] = field_descr.filler.run(user, text_normalization_result, params)
            event = Event(type=HistoryConstants.types.FIELD_EVENT,
                          scenario=self.root_id,
                          content={HistoryConstants.content_fields.FIELD: field_key},
                          result=HistoryConstants.event_results.FILLED)
            user.history.add_event(event)
        return result

    def _extract_data(self, form, text_normalization_result, user, params):
        result = {}

        callback_id = user.message.callback_id
        action_params = user.behaviors.get_callback_action_params(callback_id) or {}
        request_field = action_params.get(REQUEST_FIELD)
        if request_field and request_field["type"] == "integration":
            field = form.fields[request_field["id"]]
            field_descr = form.description.fields[request_field["id"]]
            if field.available:
                result.update(self._extract_by_field_filler(request_field["id"], field_descr, text_normalization_result,
                                                            user, params))
        else:
            for field_key, field_descr in form.description.fields.items():
                field = form.fields[field_key]
                if field.available and isinstance(field, QuestionField):
                    result.update(self._extract_by_field_filler(field_key, field_descr,
                                                                text_normalization_result, user, params))
        return result

    def _validate_extracted_data(self, user, text_preprocessing_result, form, data_extracted, params):
        error_msgs = []
        for field_key, field in form.description.fields.items():
            value = data_extracted.get(field_key)
            # is not None is necessary, because 0 and False should be checked, None - shouldn't fill
            if value is not None and not field.field_validator.requirement.check(value, params):
                log_params = {
                    log_const.KEY_NAME: log_const.SCENARIO_RESULT_VALUE,
                    "field_key": field_key
                }
                message = "Field is not valid: %(field_key)s"
                log(message, user, log_params)
                actions = field.field_validator.actions
                error_msgs = self.get_action_results(user, text_preprocessing_result, actions)
                break
        return error_msgs

    def _fill_form(self, user, text_preprocessing_result, form, data_extracted):
        on_filled_actions = []
        fields = form.fields
        scenario_model = user.scenario_models[self.id]
        is_break = False
        for description in fields:
            key = description.id
            value = data_extracted.get(key)
            field = fields[key]
            if field.fill(value):
                _action = self.get_action_results(user=user, text_preprocessing_result=text_preprocessing_result,
                                                  actions=field.description.on_filled_actions)
                on_filled_actions.extend(_action)
                if scenario_model.break_scenario:
                    is_break = True
                    return _action, is_break
        return on_filled_actions, is_break

    def get_reply(self, user, text_preprocessing_result, reply_actions, field, form):
        action_params = {}
        if field:
            field.set_available()
            actions = field.description.requests
            params = {
                log_const.KEY_NAME: log_const.SCENARIO_RESULT_VALUE,
                "field": field.description.id
            }
            message = "Ask question on field: %(field)s"
            log(message, user, params)
            action_params[REQUEST_FIELD] = {"type": field.description.type, "id": field.description.id}
            action_messages = self.get_action_results(user, text_preprocessing_result, actions, action_params)
        else:
            actions = reply_actions
            params = {
                log_const.KEY_NAME: log_const.SCENARIO_RESULT_VALUE,
                "id": self.id
            }
            message = "Finished scenario: %(id)s"
            log(message, user, params)
            user.preprocessing_messages_for_scenarios.clear()
            action_messages = self.get_action_results(user, text_preprocessing_result, actions, action_params)
            user.last_scenarios.delete(self.id)
        return action_messages

    @monitoring.got_histogram("scenario_time")
    def run(self, text_preprocessing_result, user, params: Dict[str, Any] = None):
        form = self._get_form(user)
        user.last_scenarios.add(self.id, text_preprocessing_result)
        user.preprocessing_messages_for_scenarios.add(text_preprocessing_result)

        data_extracted = self._extract_data(form, text_preprocessing_result, user, params)
        logging_params = {"data_extracted_str": str(data_extracted)}
        logging_params.update(self._log_params())
        log("Extracted data=%(data_extracted_str)s", user, logging_params)

        validation_error_msg = self._validate_extracted_data(user, text_preprocessing_result, form, data_extracted, params)
        if validation_error_msg:
            reply_messages = validation_error_msg
        else:
            reply_messages, is_break = self._fill_form(user, text_preprocessing_result, form, data_extracted)
            if not is_break:
                field = self._field(form, text_preprocessing_result, user, params)
                if field:
                    user.history.add_event(
                        Event(type=HistoryConstants.types.FIELD_EVENT,
                              scenario=self.root_id,
                              content={HistoryConstants.content_fields.FIELD: field.description.id},
                              result=HistoryConstants.event_results.ASK_QUESTION))
                reply = self.get_reply(user, text_preprocessing_result, self.actions, field, form)
                reply_messages.extend(reply)

        if not reply_messages:
            reply_messages = self.get_no_commands_action(user, text_preprocessing_result)

        return reply_messages



from functools import cached_property

from core.names import field


class AppInfo:

    def __init__(self, value):
        self._value = value

    @cached_property
    def project_id(self):
        return self._value.get(field.PROJECT_ID)

    @cached_property
    def system_name(self):
        return self._value.get(field.SYSTEM_NAME)

    @cached_property
    def application_id(self):
        return self._value.get(field.APPLICATION_ID)

    @cached_property
    def app_version_id(self):
        return self._value.get(field.APP_VERSION_ID)

    @cached_property
    def frontend_endpoint(self):
        return self._value.get(field.FRONTEND_ENDPOINT)

    @cached_property
    def frontend_type(self):
        return self._value.get(field.FRONTEND_TYPE)



import os
from unittest import TestCase
from unittest.mock import Mock, patch

import smart_kit
from core.text_preprocessing.preprocessing_result import TextPreprocessingResult
from scenarios.scenario_models.field.field_filler_description import (
    ApproveFiller,
    ApproveRawTextFiller
)
from smart_kit.text_preprocessing.local_text_normalizer import LocalTextNormalizer


def patch_get_app_config(mock_get_app_config):
    result = Mock()
    sk_path = os.path.dirname(smart_kit.__file__)
    result.STATIC_PATH = os.path.join(sk_path, 'template/static')
    mock_get_app_config.return_value = result
    result.NORMALIZER = LocalTextNormalizer()
    mock_get_app_config.return_value = result


class TestApproveFiller(TestCase):

    @patch('smart_kit.configs.get_app_config')
    def test_1(self, mock_get_app_config):
        patch_get_app_config(mock_get_app_config)
        items = {
            'yes_words': [
                'да',
                'конечно',
                'давай',
                'ага',
                'хорошо',
                'именно',
                'можно'
            ],
            'no_words': [
                'нет',
                'не',
                'ни',
                'нельзя',
                'отнюдь',
                'нету'
            ]
        }
        filler = ApproveFiller(items)
        normalizer = LocalTextNormalizer()

        user_phrase = 'даю'
        text_pre_result = TextPreprocessingResult(normalizer(user_phrase))
        result = filler.extract(text_pre_result, None)
        self.assertTrue(result)

        user_phrase = 'да нет'
        text_pre_result = TextPreprocessingResult(normalizer(user_phrase))
        result = filler.extract(text_pre_result, None)
        self.assertFalse(result)

        user_phrase = 'даю добро'
        text_pre_result = TextPreprocessingResult(normalizer(user_phrase))
        result = filler.extract(text_pre_result, None)
        self.assertIsNone(result)


class TestApproveRawTextFiller(TestCase):

    @patch('smart_kit.configs.get_app_config')
    def test_1(self, mock_get_app_config):
        patch_get_app_config(mock_get_app_config)
        items = {
            'yes_words': [
                'да',
                'конечно',
                'давай',
                'ага',
                'хорошо',
                'именно',
                'можно'
            ],
            'no_words': [
                'нет',
                'не',
                'ни',
                'нельзя',
                'отнюдь',
                'нету'
            ]
        }
        filler = ApproveRawTextFiller(items)
        normalizer = LocalTextNormalizer()

        user_phrase = 'конечно'
        text_pre_result = TextPreprocessingResult(normalizer(user_phrase))
        result = filler.extract(text_pre_result, None)
        self.assertTrue(result)

        user_phrase = 'да нет'
        text_pre_result = TextPreprocessingResult(normalizer(user_phrase))
        result = filler.extract(text_pre_result, None)
        self.assertFalse(result)

        user_phrase = 'даю'
        text_pre_result = TextPreprocessingResult(normalizer(user_phrase))
        result = filler.extract(text_pre_result, None)
        self.assertIsNone(result)



# coding: utf-8
import unittest

from core.model.heapq.heapq_storage import HeapqKV
from core.model.heapq.heapq_storage import heapq


class ModelsTest2(unittest.TestCase):
    def setUp(self):
        self.test_value = 10
        self.test_key_func = lambda x: x**2

    def test_heapq_storage_init(self):
        obj = HeapqKV(self.test_key_func)
        self.assertTrue(obj._heapq == [])
        self.assertTrue(obj._to_remove == set())
        self.assertTrue(obj._value_to_key == self.test_key_func)

    def test_heapq_storage_value_to_key(self):
        obj = HeapqKV(self.test_key_func)
        self.assertTrue(obj.value_to_key(self.test_value) == 100)

    def test_heapq_storage_push(self):
        obj = HeapqKV(self.test_key_func)
        obj.push(1, 2)
        self.assertTrue(obj._heapq == [(1, 2)])
        obj.push(3, 4)
        self.assertTrue(obj._heapq == [(1, 2), (3, 4)])

    def test_heapq_storage_check_key_valid(self):
        obj = HeapqKV(self.test_key_func)
        key_val = obj.value_to_key(self.test_value)
        self.assertTrue(obj._check_key_valid(self.test_value))
        obj._to_remove.add(key_val)
        self.assertTrue(not obj._check_key_valid(self.test_value))
        self.assertTrue(obj._to_remove == set())

    def test_heapq_storage_remove(self):
        obj = HeapqKV(self.test_key_func)
        self.assertTrue(obj._to_remove == set())
        obj.remove(self.test_value)
        self.assertTrue(obj._to_remove == {self.test_value})

    def test_heapq_storage_get_head_key(self):
        # while используется как аналог bool(list)

        obj = HeapqKV(self.test_key_func)
        # проверка при выполнении условия
        obj.push(2, 4)
        self.assertTrue(obj.get_head_key() == 2)
        heapq.heappop(obj._heapq)
        obj.push(3, 9)
        self.assertTrue(obj.get_head_key() == 3)
        heapq.heappop(obj._heapq)
        # функция может ничего не возвращать
        self.assertIsNone(obj.get_head_key())

        # проверка при невыполнении условия
        obj.push(2, 4)
        obj._to_remove.add(16)  # использовалась функция вовзедения в квадрат числа 4**2 = 16
        self.assertIsNone(obj.get_head_key())
        self.assertTrue(obj._heapq == [])

    def test_heapq_storage_pop(self):
        # while используется как аналог bool(list) и не даёт возникнуть исключению IndexError для pop
        obj = HeapqKV(self.test_key_func)
        # проверка при выполнении условия с возвратом
        obj.push(2, 4)
        self.assertTrue(obj.pop() == (2, 4))
        self.assertIsNone(obj.pop())
        # проверка при невыполнении условия
        obj.push(2, 4)
        obj._to_remove.add(16)
        self.assertIsNone(obj.pop())
        self.assertTrue(obj._heapq == [])



# coding: utf-8
import unittest
from unittest.mock import Mock

from smart_kit.handlers import handler_text
from smart_kit.utils.picklable_mock import PicklableMock


class HandlerTest5(unittest.TestCase):
    def setUp(self):
        self.app_name = "TestAppName"
        self.test_dialog_manager1 = Mock('dialog_manager')
        self.test_dialog_manager1.run = lambda x, y: (["TestAnswer"], True)
        self.test_dialog_manager2 = Mock('dialog_manager')
        self.test_dialog_manager2.run = lambda x, y: ("", False)
        self.test_text_preprocessing_result = Mock('text_preprocessing_result')
        self.test_text_preprocessing_result.raw = 'any raw'
        self.test_user = Mock('User')
        self.test_user.id = "1234-5678-9012"
        self.test_user.message = Mock('Message')
        self.test_user.message.incremental_id = "123456"
        self.test_user.message.logging_uuid = "1234-5678"
        self.test_user.message.message_name = "test"
        self.test_user.message.device = Mock("device")
        self.test_user.message.device.surface = "test_surface"
        self.test_user.message.channel = "test_channel"
        self.test_user.message.app_info = PicklableMock()
        self.test_user.message.app_info.project_id = "1111-1111-1111-1111"
        self.test_user.message.app_info.system_name = "test"
        self.test_payload = {'message': {1: 1}}

    def test_handler_text_init(self):
        obj1 = handler_text.HandlerText(self.app_name, None)
        obj2 = handler_text.HandlerText(self.app_name, self.test_dialog_manager1)
        self.assertIsNotNone(obj1.KAFKA_KEY)
        self.assertIsNone(obj1.dialogue_manager)
        self.assertIsNotNone(obj2.dialogue_manager)
        self.assertIsNotNone(handler_text.log_const.KEY_NAME)
        self.assertIsNotNone(handler_text.log_const.STARTUP_VALUE)
        self.assertIsNotNone(obj1.__class__.__name__)

    def test_handler_text_handle_base(self):
        obj1 = handler_text.HandlerText(self.app_name, self.test_dialog_manager1)
        obj2 = handler_text.HandlerText(self.app_name, self.test_dialog_manager2)
        self.assertTrue(obj1._handle_base(self.test_text_preprocessing_result, self.test_user) == ["TestAnswer"])
        self.assertTrue(obj2._handle_base(self.test_text_preprocessing_result, self.test_user) == [])

    def test_handler_text_run(self):
        self.assertIsNotNone(handler_text.log_const.NORMALIZED_TEXT_VALUE)
        obj1 = handler_text.HandlerText(self.app_name, self.test_dialog_manager1)
        obj2 = handler_text.HandlerText(self.app_name, self.test_dialog_manager2)
        self.assertTrue(obj1.run(self.test_payload, self.test_user) == ["TestAnswer"])
        self.assertTrue(obj2.run(self.test_payload, self.test_user) == [])



# coding: utf-8
from functools import cached_property

from core.basic_models.actions.basic_actions import Action
from core.basic_models.requirement.basic_requirements import Requirement
from core.model.factory import factory, list_factory


class TreeScenarioNode:
    def __init__(self, items, id):
        items = items or {}
        self.id = id
        self.form_key = items["form_key"]
        self._requirement = items.get("requirement")
        self._actions = items.get("actions")
        self.available_nodes = items.get("available_nodes")

    @cached_property
    @factory(Requirement)
    def requirement(self):
        return self._requirement

    @cached_property
    @list_factory(Action)
    def actions(self):
        return self._actions



import core.logging.logger_constants as log_const
from core.logging.logger_utils import log
from core.repositories.shard_repository import ShardRepository


class FolderRepository(ShardRepository):

    def __init__(self, path, loader, source=None, *args, **kwargs):
        super(FolderRepository, self).__init__(path, loader, source, *args, **kwargs)

    def _load_item(self, name):
        try:
            with self.source.open(name, mode='rb') as shard_stream:
                shard_binary_data = shard_stream.read()
                shard_data = shard_binary_data.decode()
                loaded_data = self.loader(shard_data)
        except:
            params = {
                "error_repository_path": str(self.path),
                log_const.KEY_NAME: log_const.EXCEPTION_VALUE
            }
            log("FolderRepository.__load_item loading failed with file %(error_repository_path)s",
                          params=params,
                          level="WARNING")
            raise
        return loaded_data

    def _form_file_upload_map(self, shard_desc):
        filename_to_data = {}
        for shard in shard_desc:
            shard_data = self._load_item(shard)
            if shard_data:
                filename_to_data.update({shard: shard_data})
        return filename_to_data

    def load(self):
        shard_desc = self.get_shard_desc()
        self.fill(self._form_file_upload_map(shard_desc))
        super(FolderRepository, self).load()

    def load_in_parts(self, count):
        self.clear()
        shard_desc = self.get_shard_desc()
        for i in range(0, len(shard_desc), count):
            desc_slice = shard_desc[i: i + count]
            self.fill_on_top(self._form_file_upload_map(desc_slice))
            log("FolderRepository.load_in_parts start loading [%(current_count)s/%(all_count)s]",
                          params={"current_count": i + len(desc_slice),
                                  "all_count": len(shard_desc)},
                          level="WARNING")
            super(FolderRepository, self).load()

    def get_shard_desc(self):
        shard_desc = self.source.list_dir(self.path)
        if len(shard_desc) == 0:
            params = {
                "error_repository_path": self.path,
                log_const.KEY_NAME: log_const.EXCEPTION_VALUE
            }
            log("FolderRepository.load got empty or nonexistent folder path %(error_repository_path)s",
                          params=params, level="WARNING")
        return shard_desc

    def check_load_in_parts(self):
        return True



# coding: utf-8
import random
from copy import copy
from functools import cached_property
from itertools import chain
from typing import Union, Dict, List, Any, Optional, Tuple, TypeVar, Type

from core.basic_models.actions.basic_actions import CommandAction
from core.basic_models.actions.command import Command
from core.basic_models.answer_items.answer_items import SdkAnswerItem
from core.model.base_user import BaseUser
from core.model.factory import list_factory
from core.text_preprocessing.base import BaseTextPreprocessingResult
from core.unified_template.unified_template import UnifiedTemplate, UNIFIED_TEMPLATE_TYPE_NAME

T = TypeVar("T")

ANSWER_TO_USER = "ANSWER_TO_USER"


class NodeAction(CommandAction):
    """ NodeAction расширяет функционал CommandAction, добавляя рендеринг Jinja-шаблонов внутри полей "nodes" и
    "support_templates".

    """
    version: Optional[int]
    command: str
    nodes: Dict[str, List[List[str]]]
    support_templates: Dict[str, Any]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super(NodeAction, self).__init__(items, id)
        items = items or {}
        self._nodes = items.get("nodes") or {}
        self._support_templates = items.get("support_templates") or {}
        self.no_empty_nodes = items.get("no_empty_nodes", False)

    @cached_property
    def nodes(self) -> Dict[str, Union[str, T]]:
        return {k: self._get_template_tree(t) for k, t in self._nodes.items()}

    @cached_property
    def support_templates(self) -> Dict[str, Union[str, T]]:
        return {k: self._get_template_tree(t) for k, t in self._support_templates.items()}

    def _get_template_tree(self, value: Union[str, T]) -> T:
        is_dict_unified_template = isinstance(value, dict) and value.get("type") == UNIFIED_TEMPLATE_TYPE_NAME
        if isinstance(value, str) or is_dict_unified_template:
            result = UnifiedTemplate(value)
        elif isinstance(value, dict):
            result = {}
            for inner_key, inner_value in value.items():
                result[inner_key] = self._get_template_tree(inner_value)
        elif isinstance(value, list):
            result = []
            for inner_value in value:
                result.append(self._get_template_tree(inner_value))
        else:
            result = value
        return result

    def _get_rendered_tree(self, value: T, params: Dict, no_empty=False) -> Union[str, Dict, List]:
        params = copy(params)
        for support_key, support_template in self.support_templates.items():
            params[support_key] = support_template.render(params)
        return self._get_rendered_tree_recursive(value, params, no_empty=no_empty)

    def _get_rendered_tree_recursive(self, value: T, params: Dict, no_empty=False) -> Union[str, Dict, List]:
        value_type: Type[T] = type(value)
        if value_type is dict:
            result = {}
            for inner_key, inner_value in value.items():
                rendered = self._get_rendered_tree_recursive(inner_value, params, no_empty=no_empty)
                if rendered != "" or not no_empty:
                    result[inner_key] = rendered
        elif value_type is list:
            result = []
            for inner_value in value:
                rendered = self._get_rendered_tree_recursive(inner_value, params, no_empty=no_empty)
                if rendered != "" or not no_empty:
                    result.append(rendered)
        elif value_type is UnifiedTemplate:
            result = value.render(params)
        else:
            result = value
        return result

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        raise NotImplementedError


class StringAction(NodeAction):
    """ StringAction расширяет NodeAction, реализуя метод run.

    Метод run возвращает команду, составленную из заданных в теле действия типа команды, типа запроса и передаваемых
    данных, а также извлечённых из `text_preprocessing_result` параметров.

    Example:
    {
      "pay_phone_cmd": {
        "type": "string",
        "command": "recharge_mobile",
        "nodes": {
          "phone": "{% if approve and phone_number is defined and phone_number not in (None, 1) %}{{ phone_number }}{% endif %}",
          "amount": "{% if approve and amount is defined %}{{ amount | int }}{% endif%}",
          "currency": "{% if approve and currency is defined %}{{ currency }}{% endif%}",
          "card": "{% if approve and card is defined %}{{ card }}{% endif %}"
        }
      }
    }
    """
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super(StringAction, self).__init__(items, id)

    def _generate_command_context(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
                                  params: Optional[Dict[str, Union[str, float, int]]] = None) -> Dict:
        command_params = dict()
        collected = user.parametrizer.collect(text_preprocessing_result, filter_params={"command": self.command})
        params.update(collected)

        for key, value in self.nodes.items():
            rendered = self._get_rendered_tree(value, params, self.no_empty_nodes)
            if rendered != "" or not self.no_empty_nodes:
                command_params[key] = rendered
        return command_params

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        # Example: Command("ANSWER_TO_USER", {"answer": {"key1": "string1", "keyN": "stringN"}})
        params = params or {}
        command_params = self._generate_command_context(user, text_preprocessing_result, params)
        commands = [Command(self.command, command_params, self.id, request_type=self.request_type,
                            request_data=self.request_data)]
        return commands


class AfinaAnswerAction(NodeAction):
    """
    Example:
    {
      "type": "random_field_answer",
      "nodes":
      {
          "pronounce_text": ["pronounceText1", "pronounceText2"]
      }
    }

    Output:
    [command1(pronounceText)]
    """
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super(AfinaAnswerAction, self).__init__(items, id)
        self.command: str = ANSWER_TO_USER

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        params = user.parametrizer.collect(text_preprocessing_result, filter_params={"command": self.command})
        answer_params = dict()
        result = []

        nodes = self.nodes.items() if self.nodes else []
        for key, template in nodes:
            if template:
                choice_index = random.randint(0, len(template) - 1)
                rendered = self._get_rendered_tree(template[choice_index], params, self.no_empty_nodes)
                if rendered != "" or not self.no_empty_nodes:
                    answer_params[key] = rendered

        if answer_params:
            result = [Command(self.command, answer_params, self.id, request_type=self.request_type,
                              request_data=self.request_data)]
        return result


class SDKAnswer(NodeAction):
    """
    Example:
    {
      "type": "sdk_answer",
      # можно без nodes
      "nodes":
      {
        "pronounceText": ["pronounceText1", "pronounceText2"],
        "items": [
          {
            "bubble": {
              "text": ["Text1", "Text2"]
            }
          },
          {
            "card": {
              "type": "simple_list",
              "header": "1 доллар США",
              "items": [
                {
                  "title": "Купить",
                  "body": "67.73 RUR"
                },
                {
                  "title": "Продать",
                  "body": "64.56 RUR"
                }
              ],
              "footer": "Сбербанк Онлайн на сегодня 17:53 при обмене до 1000 USD"
            }
          }
        ],
        "suggestions": {
         "buttons": [{
            "title": "Отделения",
            "action": {
              "text": ["Где ближайщие отделения сбера?"],
              "type": "text"
            }
         }]
        }
      }
    }

    Output:
    [command1(pronounceTexti, Textj, card, buttons)]
    ответ c карточками с случайным выбором текстов из массива
    карточки на андроиде требуют sdk_version не ниже "20.03.0.0"
    """
    INDEX_WILDCARD = "*index*"
    RANDOM_PATH = [['items', INDEX_WILDCARD, 'bubble', 'text'], ['pronounceText'],
                   ['suggestions', 'buttons', INDEX_WILDCARD, 'title']]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super(SDKAnswer, self).__init__(items, id)
        self.command: str = ANSWER_TO_USER
        if self._nodes == {}:
            self._nodes = {i: items.get(i) for i in items if
                           i not in ['random_paths', 'same_ans', 'type', 'support_templates', 'no_empty_nodes']}

    # функция идет по RANDOM_PATH, числа в нем считает индексами массива,
    # INDEX_WILDCARD - произвольным индексом массива, прочее - ключами словаря
    # в конце пути предполагается непустой массив
    def get_by_path(self, input_dict: Union[list, dict], nested_key: List[str]) -> List[Tuple[Union[list, dict], str]]:
        internal_dict_value = input_dict
        for ik, k in enumerate(nested_key):
            last_dict = internal_dict_value
            if k == self.INDEX_WILDCARD:
                return list(chain.from_iterable(
                    [self.get_by_path(wildcard_item, nested_key[ik + 1:]) for wildcard_item in last_dict]
                ))
            if k.isdigit():
                internal_dict_value = internal_dict_value[int(k)]
            else:
                internal_dict_value = internal_dict_value.get(k, None)
            if internal_dict_value is None:
                return []
        return [(last_dict, k)]

    def do_random(self, input_dict: Union[list, dict]):
        dicts = list(chain.from_iterable([self.get_by_path(input_dict, j) for j in self.RANDOM_PATH]))
        max_length = max(map(lambda x: len(x[0][x[1]]), dicts))
        random_index = random.randrange(max_length)
        for (d, k) in dicts:
            d[k] = d[k][random_index % len(d[k])]

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        result = []
        params = user.parametrizer.collect(text_preprocessing_result, filter_params={"command": self.command})
        rendered = self._get_rendered_tree(self.nodes, params, self.no_empty_nodes)
        self.do_random(rendered)
        if rendered or not self.no_empty_nodes:
            result = [
                Command(
                    self.command,
                    rendered,
                    self.id,
                    request_type=self.request_type,
                    request_data=self.request_data,
                )
            ]
        return result


class SDKAnswerToUser(NodeAction):
    """
    Example:
    {
        "type": "sdk_answer_to_user",
        "root":
        [
          {
            "type": "pronounce_text",
            "text": "ans"
          }
        ],
        "static": {
          "ios_card": {
            "type": "list_card",
            "cells": [
              {
                "ios_params": "ios"
              }
            ]
          },
          "android_card": {
            "type": "list_card",
            "cells": [
              {
                "android_params": "android"
              }
            ]
          },
          "tittle1": "static tittle1",
          "tittle2": "static tittle2",
          "sg_dl": "www.www.www",
          "sg_text": "static suggest text"
        },
        "random_choice": [
          {
            "ans": "random text1"
          },
          {
            "ans": "random text2"
          }
        ],
        "items": [
          {
            "type": "item_card",
            "value": "ios_card",
            "requirement": {
              "type": "external",
              "requirement": "OCTOBER_iOS"
            }
          },
          {
            "type": "item_card",
            "value": "android_card",
            "requirement": {
              "type": "external",
              "requirement": "OCTOBER_android"
            }
          },
          {
            "type": "bubble_text",
            "text": "ans"
          }
        ],
        "suggestions": [
          {
            "type": "suggest_text",
            "text": "sg_text",
            "title": "tittle1"
          },
          {
            "type": "suggest_deeplink",
            "text": "sg_text",
            "deep_link": "tittle1"
          }
        ]
    }

    Output:
    {
        "items":
            [{
                "card": {"type": "list_card", "cells": [{"ios_params": "ios"}]}},
                {"bubble": {"text": "random texti", "markdown": True}
            }],
        "suggestions":
            {"buttons":
                [
                    {"title": "static tittle1", "action": {"text": "static suggest text", "type": "text"}},
                    {"title": "static tittle2", "action": {"deep_link": "www.www.www", "type": "deep_link"}}
                ]
            },
        "pronounceText": "random texti"
    }
    ответ c карточками с случайным выбором текстов из random_choice
    карточки на андроиде требуют sdk_version не ниже "20.03.0.0"
    """
    ITEMS = "items"
    SUGGESTIONS = "suggestions"
    SUGGESTIONS_TEMPLATE = "suggestions_template"
    BUTTONS = "buttons"
    STATIC = "static"
    RANDOM_CHOICE = "random_choice"
    COMMAND = "command"
    ROOT = "root"

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super(SDKAnswerToUser, self).__init__(items, id)
        self.command: str = ANSWER_TO_USER
        self._nodes[self.STATIC] = items.get(self.STATIC, {})
        self._nodes[self.RANDOM_CHOICE] = items.get(self.RANDOM_CHOICE, {})
        self._nodes[self.SUGGESTIONS] = items.get(self.SUGGESTIONS, {})
        self._nodes[self.SUGGESTIONS_TEMPLATE] = items.get(self.SUGGESTIONS_TEMPLATE, {})
        self._items = items.get(self.ITEMS, {})
        self._suggests = items.get(self.SUGGESTIONS, {})
        self._suggests_template = items.get(self.SUGGESTIONS_TEMPLATE)
        self._root = items.get(self.ROOT, {})

        self.items: List[SdkAnswerItem] = self.build_items()
        self.suggests: List[SdkAnswerItem] = self.build_suggests()
        self.root: List[SdkAnswerItem] = self.build_root()

    @list_factory(SdkAnswerItem)
    def build_items(self):
        return self._items

    @list_factory(SdkAnswerItem)
    def build_suggests(self):
        return self._suggests

    @list_factory(SdkAnswerItem)
    def build_root(self):
        return self._root

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:

        result = []
        params = user.parametrizer.collect(text_preprocessing_result, filter_params={self.COMMAND: self.command})
        rendered = self._get_rendered_tree(self.nodes[self.STATIC], params, self.no_empty_nodes)
        if self._nodes[self.RANDOM_CHOICE]:
            random_node = random.choice(self.nodes[self.RANDOM_CHOICE])
            rendered_random = self._get_rendered_tree(random_node, params, self.no_empty_nodes)
            rendered.update(rendered_random)
        out = {}
        for item in self.items:
            if item.requirement.check(text_preprocessing_result, user, params):
                out.setdefault(self.ITEMS, []).append(item.render(rendered))

        if self._suggests_template is not None:
            out[self.SUGGESTIONS] = self._get_rendered_tree(self.nodes[self.SUGGESTIONS_TEMPLATE], params,
                                                            self.no_empty_nodes)
        else:
            for suggest in self.suggests:
                if suggest.requirement.check(text_preprocessing_result, user, params):
                    data_dict = out.setdefault(self.SUGGESTIONS, {self.BUTTONS: []})
                    buttons = data_dict[self.BUTTONS]
                    rendered_text = suggest.render(rendered)
                    buttons.append(rendered_text)
        for part in self.root:
            if part.requirement.check(text_preprocessing_result, user):
                out.update(part.render(rendered))
        if rendered or not self.no_empty_nodes:
            result = [
                Command(
                    self.command,
                    out,
                    self.id,
                    request_type=self.request_type,
                    request_data=self.request_data,
                )
            ]
        return result



import json
import typing
from collections import defaultdict
from wsgiref.simple_server import make_server

import scenarios.logging.logger_constants as log_const
from core.basic_models.actions.command import Command
from core.configs.global_constants import CALLBACK_ID_HEADER
from core.logging.logger_utils import log
from core.message.from_message import SmartAppFromMessage, basic_error_message
from core.utils.stats_timer import StatsTimer
from smart_kit.compatibility.commands import combine_commands
from smart_kit.message.smartapp_to_message import SmartAppToMessage
from smart_kit.names import message_names
from smart_kit.start_points.base_main_loop import BaseMainLoop


class BaseHttpMainLoop(BaseMainLoop):
    HEADER_START_WITH = "HTTP_SMART_APP_"
    BAD_REQUEST_COMMAND = Command(message_names.ERROR, {"code": -1, "description": "Invalid Request Message"})
    NO_ANSWER_COMMAND = Command(message_names.NOTHING_FOUND)
    BAD_ANSWER_COMMAND = Command(message_names.ERROR, {"code": -1, "description": "Invalid Answer Message"})

    def run(self):
        raise NotImplementedError

    def stop(self, signum, frame):
        raise NotImplementedError

    def handle_message(self, message: SmartAppFromMessage) -> typing.Tuple[int, str, SmartAppToMessage]:
        if not message.validate():
            result = 400, "BAD REQUEST", SmartAppToMessage(
                self.BAD_REQUEST_COMMAND,
                message=message,
                request=None,
            )
            try:
                result[2].as_dict
            except (json.JSONDecodeError, KeyError):
                result = 400, "BAD REQUEST", SmartAppToMessage(
                        self.BAD_REQUEST_COMMAND,
                        message=basic_error_message,
                        request=None,
                    )
            finally:
                return result


        answer, stats = self.process_message(message)
        if not answer:
            return 204, "NO CONTENT", SmartAppToMessage(self.NO_ANSWER_COMMAND, message=message, request=None)

        answer_message = SmartAppToMessage(
            answer, message, request=None,
            validators=self.to_msg_validators)
        if answer_message.validate():
            return 200, "OK", answer_message
        else:
            return 500, "BAD ANSWER", SmartAppToMessage(self.BAD_ANSWER_COMMAND, message=message, request=None)

    def process_message(self, message: SmartAppFromMessage, *args, **kwargs):
        stats = ""
        log("INCOMING DATA: %(masked_message)s",
            params={
                log_const.KEY_NAME: "incoming_policy_message",
                "masked_message": message.masked_value,
                "message_id": message.incremental_id,
            })

        db_uid = message.db_uid
        with StatsTimer() as load_timer:
            user = self.load_user(db_uid, message)
        stats += "Loading time: {} msecs\n".format(load_timer.msecs)
        with StatsTimer() as script_timer:
            commands = self.model.answer(message, user)
            if commands:
                answer = self._generate_answers(user, commands, message)
            else:
                answer = None

        stats += "Script time: {} msecs\n".format(script_timer.msecs)
        with StatsTimer() as save_timer:
            self.save_user(db_uid, user, message)
        stats += "Saving time: {} msecs\n".format(save_timer.msecs)
        log(stats, user=user, params={log_const.KEY_NAME: "timings"})
        self.postprocessor.postprocess(user, message)
        return answer, stats

    def _get_headers(self, environ):
        return [(key, value) for key, value in environ.items() if key.startswith(self.HEADER_START_WITH)]

    # noinspection PyMethodMayBeStatic
    def _get_outgoing_headers(self, incoming_headers, command=None):
        headers = {"CONTENT_TYPE": "application/json"}
        headers.update(incoming_headers)

        if command:
            callback_id = command.request_data.get(CALLBACK_ID_HEADER)
            if callback_id:
                headers[CALLBACK_ID_HEADER] = callback_id

        return list(headers.items())

    def _generate_answers(self, user, commands, message, **kwargs):
        commands = combine_commands(commands, user)
        if len(commands) > 1:
            raise ValueError
        answer = commands.pop() if commands else None

        return answer


class HttpMainLoop(BaseHttpMainLoop):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._server = None
        self.location_maper = defaultdict(lambda: self.iterate)
        self.location_maper["/health"] = self.get_health_check

    def __call__(self, environ, start_response):
        location = environ.get('PATH_INFO')
        location_handler = self.location_maper[location]
        return location_handler(environ, start_response)

    def get_health_check(self, environ, start_response):
        status, reason, answer = 200, "OK", "ok"
        start_response(f"{status} {reason}", self._get_outgoing_headers({}))
        return [answer.encode()]

    def iterate(self, environ, start_response):
        try:
            content_length = int(environ.get('CONTENT_LENGTH', '0'))
            body = environ["wsgi.input"].read(content_length).decode()
            headers = self._get_headers(environ)
        except KeyError:
            log("Error in request data", level="ERROR")
            raise Exception("Error in request data")

        message = SmartAppFromMessage(json.loads(body), headers=headers, headers_required=False,
                                      validators=self.from_msg_validators)

        status, reason, answer = self.handle_message(message)

        start_response(f"{status} {reason}", self._get_outgoing_headers(headers, answer.command))
        return [answer.value.encode()]

    def run(self):
        self._server = make_server('0.0.0.0', 8000, self.iterate)
        log(
            '''
                Application start via "python manage.py run_app" recommended only for local testing. 
                For production it is recommended to start using "gunicorn --config wsgi_config.py 'wsgi:create_app()'
            ''',
            level="WARNING")
        self._server.serve_forever()

    def stop(self, signum, frame):
        if self._server:
            self._server.server_close()
        exit(0)



import collections
import json
import operator
import re
from functools import cached_property
from itertools import islice
from typing import Dict, List, Union, Optional, Any, Callable, Pattern, Set

from jinja2 import exceptions as jexcept

import core.basic_models.classifiers.classifiers_constants as cls_const
import core.logging.logger_constants as core_log_const
import scenarios.logging.logger_constants as log_const
from core.basic_models.classifiers.basic_classifiers import Classifier, ExternalClassifier
from core.logging.logger_utils import log, log_classifier_result
from core.model.factory import build_factory
from core.model.factory import factory
from core.model.factory import list_factory
from core.model.registered import Registered
from core.text_preprocessing.base import BaseTextPreprocessingResult
from core.text_preprocessing.preprocessing_result import TextPreprocessingResult
from core.unified_template.unified_template import UnifiedTemplate
from core.utils.exception_handlers import exc_handler
from core.utils.pickle_copy import pickle_deepcopy
from core.utils.stats_timer import StatsTimer
from core.utils.period_determiner import extract_words_describing_period, period_determiner
from scenarios.user.user_model import User

field_filler_description = Registered()

field_filler_factory = build_factory(field_filler_description)


class FieldFillerDescription:

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        items = items or {}
        self.id = id
        self.version = items.get("version", -1)

    def _log_params(self):
        return {
            log_const.KEY_NAME: log_const.FILLER_EXTRACT_VALUE,
            "filler": self.__class__.__name__
        }

    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> None:
        return None

    def on_extract_error(self, text_preprocessing_result, user, params=None):
        log("exc_handler: Filler failed to extract. Return None. MESSAGE: %(masked_message)s.", user,
            {log_const.KEY_NAME: core_log_const.HANDLED_EXCEPTION_VALUE, "masked_message": user.message.masked_value},
            level="ERROR", exc_info=True)
        return None

    def run(self, user: User, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Any]] = None) -> None:
        return self.extract(text_preprocessing_result, user, params)

    def _postprocessing(self, user: User, item: str) -> None:
        last_scenario_name = user.last_scenarios.last_scenario_name
        user.scenario_models[last_scenario_name].postprocessing_data.append(item)


class ExternalFieldFillerDescription(FieldFillerDescription):
    filler: Optional[str]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(ExternalFieldFillerDescription, self).__init__(items, id)
        self.filler = items.get("filler")

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult,
                user: User, params: Dict[str, Any] = None) -> Optional[Union[int, float, str, bool, List, Dict]]:
        filler = user.descriptions["external_field_fillers"][self.filler]
        return filler.run(user, text_preprocessing_result, params)


class CompositeFiller(FieldFillerDescription):
    fillers: Optional[List[FieldFillerDescription]]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(CompositeFiller, self).__init__(items, id)
        self._fillers: Optional[List[Dict[str, Any]]] = items.get("fillers") or []
        self.fillers = self.build_fillers()

    @list_factory(FieldFillerDescription)
    def build_fillers(self):
        return self._fillers

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult,
                user: User, params: Dict[str, Any] = None) -> Optional[Union[int, float, str, bool, List, Dict]]:
        extracted = None
        for filler in self.fillers:
            extracted = filler.extract(text_preprocessing_result, user, params)
            if extracted is not None:
                break
        return extracted


class AvailableInfoFiller(FieldFillerDescription):
    loader: Optional[str]
    value: Union[str, Dict]
    loaders = collections.defaultdict(str, {"json": json.loads, "float": float, "int": int})

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(AvailableInfoFiller, self).__init__(items, id)
        value = items['value']
        self.loader = items.get('loader')
        self.template: UnifiedTemplate = UnifiedTemplate(value)

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult,
                user: User, params: Dict[str, Any] = None) -> Optional[Union[int, float, str, bool, List, Dict]]:
        params = params or {}
        collected = user.parametrizer.collect(text_preprocessing_result)
        params.update(collected)
        try:
            # if path is wrong, it may fail with UndefinedError
            # notion: {key: None} will return "None";
            # not existing key or value "" will return ""; otherwise question in scenario will go in cycles
            value = self.template.silent_render(params)
        except jexcept.UndefinedError:
            value = None

        if self.loader:
            if value:
                loader = self.loaders[self.loader]
                value = loader(value)
            else:
                value = None

        return value


class FirstNumberFiller(FieldFillerDescription):
    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[int]:
        numbers = text_preprocessing_result.num_token_values
        if numbers:
            log_params = self._log_params()
            log_params["numbers"] = str(numbers)
            message = "Filler: %(filler)s, Numbers: %(numbers)s"
            log(message, user, log_params)
        return numbers[0] if numbers else None


class FirstCurrencyFiller(FieldFillerDescription):

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        currencies = text_preprocessing_result.ccy_token_values
        if currencies:
            log_params = self._log_params()
            log_params["currencies"] = str(currencies)
            message = "Filler: %(filler)s, Currencies: %(currencies)s"
            log(message, user, log_params)
        return currencies[0] if currencies else None


class FirstOrgFiller(FieldFillerDescription):

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        orgs = text_preprocessing_result.org_token_values
        if orgs:
            log_params = self._log_params()
            log_params["orgs"] = str(orgs)
            message = "Filler: %(filler)s, Organisations: %(orgs)s"
            log(message, user, log_params)
        return orgs[0] if orgs else None


class FirstGeoFiller(FieldFillerDescription):

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        geos = text_preprocessing_result.geo_token_values
        if geos:
            log_params = self._log_params()
            log_params["geos"] = str(geos)
            message = "Filler: %(filler)s, Toponyms: %(geos)s"
            log(message, user, log_params)
        return geos[0] if geos else None


class RegexpFieldFiller(FieldFillerDescription):
    regexp: str
    delimiter: Optional[str]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(RegexpFieldFiller, self).__init__(items, id)
        self.regexp = items["exp"]
        self.delimiter = items.get("delimiter", ",")

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        original_text = text_preprocessing_result.original_text
        match = re.findall(self.regexp, original_text)
        if match:
            log_params = self._log_params()
            log_params["match"] = str(match)
            message = "Filler: %(filler)s, Data: %(match)s"
            log(message, user, log_params)
            return self.delimiter.join(match)


class RegexpAndStringOperationsFieldFiller(RegexpFieldFiller):
    regexp: str
    delimiter: Optional[str]
    operations: List[Dict]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(RegexpAndStringOperationsFieldFiller, self).__init__(items, id)
        self.operations = items["operations"]
        self.functions_mapping: Dict[str, Callable] = {
            "strip": str.strip,
            "rstrip": str.rstrip,
            "lstrip": str.lstrip,
            "upper": str.upper,
            "lower": str.lower
        }

    def _operation(self, original_text, typeOp, amount):
        func = self.functions_mapping[typeOp]
        return func(original_text, amount) if amount else func(original_text)

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        original_text = text_preprocessing_result.original_text
        if self.operations:
            for op in self.operations:
                original_text = self._operation(original_text, op["type"], op.get("amount"))
        text_preprocessing_result_copy = pickle_deepcopy(text_preprocessing_result)
        text_preprocessing_result_copy.original_text = original_text
        return super(RegexpAndStringOperationsFieldFiller, self).extract(text_preprocessing_result_copy, user, params)


class AllRegexpsFieldFiller(FieldFillerDescription):
    exps: Optional[List[str]]
    delimiter: Optional[str]
    original_text_lower: Optional[bool]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(AllRegexpsFieldFiller, self).__init__(items, id)
        self.exps = items.get("exps") or []
        self.regexps: List[Pattern] = [re.compile(r) for r in self.exps]
        self.delimiter = items.get("delimiter") or ","
        self.original_text_lower = items.get("original_text_lower") or False

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        original_text = text_preprocessing_result.original_text
        if self.original_text_lower:
            original_text = original_text.lower()
        matches = []
        for r in self.regexps:
            matches.extend(r.findall(original_text))
        if matches:
            result = self.delimiter.join(matches)
            log_params = self._log_params()
            log_params["regexp_result"] = str(result)
            message = "Filler: %(filler)s, Data: %(regexp_result)s"
            log(message, user, log_params)
            return result


class FirstPersonFiller(FieldFillerDescription):

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[Dict[str, str]]:
        persons = text_preprocessing_result.person_token_values
        if persons:
            log_params = self._log_params()
            log_params["persons"] = str(persons)
            message = "Filler: %(filler)s, Persons: %(persons)s"
            log(message, user, log_params)
        return persons[0] if persons else None


class PreviousMessagesFiller(FieldFillerDescription):
    filler: Optional[FieldFillerDescription]
    count: Optional[int]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(PreviousMessagesFiller, self).__init__(items, id)
        self._filler: Optional[Dict[str, Any]] = items.get("filler")
        self.filler = self.build_filler()
        self.count = items.get("count")

    @factory(FieldFillerDescription)
    def build_filler(self):
        return self._filler

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        result = self.filler.extract(text_preprocessing_result, user, params)
        if result is None:
            result = self._try_extract_last_messages(user, params)
        return result

    def _try_extract_last_messages(self, user, params):
        processed_items = user.preprocessing_messages_for_scenarios.processed_items
        count = self.count - 1 if self.count else len(processed_items)
        for preprocessing_result_raw in islice(processed_items, 0, count):
            preprocessing_result = TextPreprocessingResult(preprocessing_result_raw)
            result = self.filler.extract(preprocessing_result, user, params)
            if result is not None:
                return result


class UserIdFiller(FieldFillerDescription):

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        result = user.message.uuid.get('userId')
        return result


class IntersectionFieldFiller(FieldFillerDescription):
    cases: Optional[Dict[str, List[str]]]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(IntersectionFieldFiller, self).__init__(items, id)
        self.cases = items.get("cases") or {}
        self.default = items.get("default")
        if items.get("strict"):
            self.operator = operator.eq
        else:
            self.operator = operator.ge

        from smart_kit.configs import get_app_config
        app_config = get_app_config()

        self.normalized_cases = []
        for key, val in self.cases.items():
            tokens_list = []
            for message in app_config.NORMALIZER.normalize_sequence(val):
                case = set()
                for norm in message["tokenized_elements_list"]:
                    if norm.get("token_type") != "SENTENCE_ENDPOINT_TOKEN":
                        case.add(norm.get("lemma"))
                tokens_list.append(case)
            self.normalized_cases.append((key, tokens_list))

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: TextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        tpr_tokenized_set = {norm.get("lemma") for norm in text_preprocessing_result.tokenized_elements_list_pymorphy if
                             norm.get("token_type") != "SENTENCE_ENDPOINT_TOKEN"}
        for key, tokens_list in self.normalized_cases:
            for tokens in tokens_list:
                if self.operator(tpr_tokenized_set, tokens):
                    log_params = self._log_params()
                    log_params["words_tokenized_set"] = str(tpr_tokenized_set)
                    log_params["tokens"] = str(tokens)
                    message = "Filler: %(filler)s, words_normalized_set: %(words_tokenized_set)s, tokens: %(tokens)s"
                    log(message, user, log_params)
                    return key
        if self.default:
            return self.default


class DatePeriodFiller(FieldFillerDescription):
    """
    usage:
      "вытащить_период_из_сообщения": {
        "fields": {
          "period": {
            "required": true,
            "available": true,
            "filler": {
              "type": "date_period_filler",
              "max_days_in_period": 365,
              "future_days_allowed": false
            }
          }
        }
      },
    """

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(DatePeriodFiller, self).__init__(items, id)
        self.max_days_in_period = items.get('max_days_in_period', None)
        self.future_days_allowed = items.get('future_days_allowed', False)

    def extract(self, text_preprocessing_result: TextPreprocessingResult, user: User,
                params: Optional[Dict[str, Union[str, float, int]]] = None) -> Dict[str, str]:
        if text_preprocessing_result\
            .words_tokenized_set\
            .intersection(
                [
                    'TIME_DATE_TOKEN',
                    'TIME_DATE_INTERVAL_TOKEN',
                    'PERIOD_TOKEN'
                ]):
            words_from_intent: List[Optional[str]] = text_preprocessing_result.human_normalized_text.lower().split()
        else:
            words_from_intent: List[Optional[str]] = text_preprocessing_result.original_text.lower().split()

        words_to_process: List[Optional[str]] = extract_words_describing_period(words_from_intent)
        begin_str, end_str = period_determiner(words_to_process, self.max_days_in_period, self.future_days_allowed)

        is_determined: bool = False
        is_error: bool = False
        if not (begin_str == '' or begin_str == 'error'
            or end_str == '' or end_str == 'error'):
            is_determined = True

        if begin_str == 'error' or end_str == 'error':
            is_error = True

        return {
            'date_period__is_determined': is_determined,
            'date_period__is_error': is_error,
            'date_period__begin_date': begin_str,
            'date_period__end_date': end_str
        }


class IntersectionOriginalTextFiller(FieldFillerDescription):

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(IntersectionOriginalTextFiller, self).__init__(items, id)
        self.cases = items.get("cases", {})
        self.original_cases = [(key, [{*phrase.split()} for phrase in val]) for key, val in self.cases.items()]

        self._exceptions = items.get("exceptions", {})
        self.exceptions = {key: [{*phrase.split()} for phrase in val] for key, val in self._exceptions.items()}

    def _check_exceptions(self, key, tpr_original_set):
        if key in self.exceptions.keys():
            for exc_tokens in self.exceptions[key]:
                if tpr_original_set >= exc_tokens:
                    return True
        return False

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: TextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        tpr_original_set = {*text_preprocessing_result.original_text.split()}
        for key, tokens_list in self.original_cases:
            for tokens in tokens_list:
                if tpr_original_set >= tokens and not self._check_exceptions(key, tpr_original_set):
                    log_params = self._log_params()
                    log_params["tpr_original_set"] = str(tpr_original_set)
                    log_params["tokens"] = str(tokens)
                    message = "Filler: %(filler)s, tpr_original_set: %(tpr_original_set)s, tokens: %(tokens)s"
                    log(message, user, log_params)
                    return key


class ApproveFiller(FieldFillerDescription):
    yes_words: Optional[List]
    no_words: Optional[List]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(ApproveFiller, self).__init__(items, id)

        from smart_kit.configs import get_app_config
        app_config = get_app_config()

        self.yes_words = items.get("yes_words")
        self.no_words = items.get("no_words")
        self.set_yes_words: Set = set(self.yes_words or [])
        self.set_no_words: Set = set(self.no_words or [])
        self.yes_words_normalized: Set = {
            TextPreprocessingResult(result).tokenized_string for result in
            app_config.NORMALIZER.normalize_sequence(list(self.set_yes_words))
        }
        self.no_words_normalized: Set = {
            TextPreprocessingResult(result).tokenized_string for result in
            app_config.NORMALIZER.normalize_sequence(list(self.set_no_words))
        }

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: TextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[bool]:
        if text_preprocessing_result.tokenized_string in self.yes_words_normalized:
            params = self._log_params()
            params["tokenized_string"] = text_preprocessing_result.tokenized_string
            params["yes_words"] = self.yes_words_normalized
            message = "Filler: %(filler)s, normalized_text: %(tokenized_string)s, self.yes_words: %(yes_words)s"
            log(message, user, params)
            response = True
        elif text_preprocessing_result.words_tokenized_set.intersection(self.no_words_normalized):
            params = self._log_params()
            params["tokenized_string"] = text_preprocessing_result.words_tokenized_set
            params["no_words"] = self.no_words_normalized
            message = "Filler: %(filler)s, normalized_text: %(tokenized_string)s, self.no_words: %(no_words)s"
            log(message, user, params)
            response = False
        else:
            response = None
        return response


class ApproveRawTextFiller(ApproveFiller):
    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(
            self, text_preprocessing_result: TextPreprocessingResult, user: User, params: Dict[str, Any] = None
    ) -> Optional[bool]:
        original_text = ' '.join(text_preprocessing_result.original_text.split()).lower().rstrip('!.)')
        if original_text in self.set_yes_words:
            params = self._log_params()
            params["original_text"] = original_text
            params["yes_words"] = self.set_yes_words
            message = "Filler: %(filler)s, original_text: %(original_text)s, self.yes_words: %(yes_words)s"
            log(message, user, params)
            response = True
        elif text_preprocessing_result.words_tokenized_set.intersection(self.no_words_normalized):
            params = self._log_params()
            params["words_tokenized_set"] = text_preprocessing_result.words_tokenized_set
            params["no_words"] = self.set_no_words
            message = "Filler: %(filler)s, words_normalized_set: %(words_tokenized_set)s, self.no_words: %(no_words)s"
            log(message, user, params)
            response = False
        else:
            response = None
        return response


class ClassifierFiller(FieldFillerDescription):
    """Заполняет поле одним из возможных значений, которое выдает модель классификации.
    Запрос клиента проходит классификацию на основе внешнего классификатора, после чего в качестве ответа
    берётся класс с максимальной вероятностью.
    """

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(ClassifierFiller, self).__init__(items, id)
        self._classifier = items["classifier"]
        self._cls_const_answer_key = cls_const.ANSWER_KEY

    @cached_property
    def classifier(self) -> Classifier:
        return ExternalClassifier(self._classifier)

    def _get_result(self, answers: List[Dict[str, Union[str, float, bool]]]) -> str:
        return answers[0][self._cls_const_answer_key]

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Union[str, None, List[Dict[str, Union[str, float, bool]]]]:
        result = None
        classifier = self.classifier
        with StatsTimer() as timer:
            classification_res = classifier.find_best_answer(
                text_preprocessing_result, scenario_classifiers=user.descriptions["external_classifiers"])

        log_classifier_result(classification_res, user, classifier, timer)

        if classification_res:
            params = self._log_params()
            params["answers"] = classification_res
            message = "Filler: %(filler)s, answers: %(answers)s, "
            log(message, user, params)
            result = self._get_result(classification_res)

        return result


class ClassifierFillerMeta(ClassifierFiller):

    def _get_result(self, answers: List[Dict[str, Union[str, float, bool]]]) -> List[
        Dict[str, Union[str, float, bool]]]:
        return answers



# coding: utf-8
from typing import Dict, Any, List, Union, Optional

import core.logging.logger_constants as log_const
import scenarios.logging.logger_constants as scenarios_log_const
from core.basic_models.actions.basic_actions import Action
from core.basic_models.actions.command import Command
from core.basic_models.requirement.basic_requirements import Requirement
from core.logging.logger_utils import log
from core.model.base_user import BaseUser
from core.model.factory import build_factory, list_factory
from core.model.factory import factory
from core.model.registered import Registered
from core.text_preprocessing.base import BaseTextPreprocessingResult

scenarios = Registered()

scenario_factory = build_factory(scenarios)


class BaseScenario:
    EMPTY_ANSWER_KEY = "default_empty_answer"

    scenario_description: str
    switched_off: bool
    version: int
    empty_answer: Action
    actions: List[Action]
    available_requirement: Requirement

    def __init__(self, items, id):
        self._actions = items.get("actions")
        self._available_requirement = items.get("availabe_requirement")
        self._default_empty_answer = {"type": "external", "action": self.EMPTY_ANSWER_KEY}
        self._empty_answer = items.get("empty_answer", self._default_empty_answer)
        self.id = id
        self.root_id = items.get("root_id", self.id)
        self.scenario_description = items.get("scenario_description", "")
        self.switched_off = items.get("switched_off", False)
        self.version = items.get("version", -1)

        self.empty_answer = self.build_empty_answer()
        self.actions = self.build_actions()
        self.available_requirement = self.build_available_requirement()

    @factory(Action)
    def build_empty_answer(self):
        return self._empty_answer

    @list_factory(Action)
    def build_actions(self):
        return self._actions

    @factory(Requirement)
    def build_available_requirement(self):
        return self._available_requirement

    def check_available(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser) -> bool:
        if not self.switched_off:
            return self.available_requirement.check(text_preprocessing_result, user)
        return False

    def _log_params(self) -> Dict[str, str]:
        return {log_const.KEY_NAME: log_const.SCENARIO_VALUE}

    def text_fits(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser) -> bool:
        return False

    def get_no_commands_action(self, user, text_preprocessing_result,
                               params: Dict[str, Any] = None) -> List[Command]:
        log_params = {log_const.KEY_NAME: scenarios_log_const.CHOSEN_ACTION_VALUE,
                      scenarios_log_const.CHOSEN_ACTION_VALUE: self._empty_answer}
        log(scenarios_log_const.CHOSEN_ACTION_MESSAGE, user, log_params)
        try:
            empty_answer = self.empty_answer.run(user, text_preprocessing_result, params) or []
        except KeyError:
            log_params = {log_const.KEY_NAME: scenarios_log_const.CHOSEN_ACTION_VALUE}
            log("Scenario has empty answer, but empty_answer action isn't defined",
                params=log_params, level='WARNING')
            empty_answer = []
        return empty_answer

    def get_action_results(self, user, text_preprocessing_result,
                           actions: List[Action], params: Dict[str, Any] = None) -> List[Command]:
        results = []
        for action in actions:
            result = action.run(user, text_preprocessing_result, params)
            log_params = self._log_params()
            log_params["class"] = action.__class__.__name__
            log("called action: %(class)s", user, log_params)

            if result:
                for command in result:
                    if command.action_id:
                        log_params = self._log_params()
                        log_params["id"] = command.action_id
                        log("external action id: %(id)s", user, log_params)

                    log_params = self._log_params()
                    log_params["name"] = command.name
                    log("action result name: %(name)s", user, log_params)
                results.extend(result)
        return results

    @property
    def history(self):
        return {"scenario_path": [{"scenario": self.id, "node": None}]}

    def run(self, text_preprocessing_result, user, params: Dict[str, Any] = None) -> List[Command]:
        return self.get_action_results(user, text_preprocessing_result, self.actions, params)



# coding: utf-8
import unittest

from smart_kit.models import dialogue_manager
from smart_kit.utils.picklable_mock import PicklableMock


class TestScenarioDesc(dict):
    def get_keys(self):
        return self.keys()


class ModelsTest1(unittest.TestCase):
    def setUp(self):
        self.test_user1 = PicklableMock()
        self.test_user1.name = "TestName"
        self.test_user1.last_scenarios = PicklableMock()
        self.test_user1.last_scenarios.scenarios_names = []
        self.test_user1.message.payload = {"skillId": 1, "intent": 2}
        self.test_user1.descriptions = {'external_actions': {}}
        self.test_user2 = PicklableMock()
        self.test_user2.name = "TestName"
        self.test_user2.last_scenarios = PicklableMock()
        self.test_user2.last_scenarios.scenarios_names = [1, 2]
        self.test_user2.message.payload = {"skillId": 1, "intent": 2}
        self.test_user2.descriptions = {'external_actions': {}}
        self.test_user3 = PicklableMock()
        self.test_user3.name = "TestName"
        self.test_user3.last_scenarios = PicklableMock()
        self.test_user3.last_scenarios.scenarios_names = [1]
        self.test_user3.message.payload = {"skillId": 1, "intent": 2}
        self.test_user3.descriptions = {'external_actions': {}}
        self.test_user4 = PicklableMock()
        self.test_user4.name = "TestName"
        self.test_user4.last_scenarios = PicklableMock()
        self.test_user4.last_scenarios.scenarios_names = []
        self.test_user4.message.payload = {"skillId": 1, "intent": 2}
        self.test_user4.descriptions = {'external_actions': {'before_action': PicklableMock()}}
        self.test_text_preprocessing_result = PicklableMock()
        self.test_text_preprocessing_result.name = "Result"
        self.test_scenario1 = PicklableMock()
        self.test_scenario1.scenario_description = "This is test scenario 1 desc"
        self.test_scenario1.text_fits = lambda x, y: False
        self.test_scenario1.run = lambda x, y: x.name + y.name
        self.test_scenario2 = PicklableMock()
        self.test_scenario2.scenario_description = "This is test scenario 2 desc"
        self.test_scenario2.text_fits = lambda x, y: True
        self.test_scenario2.run = lambda x, y: y.name + x.name
        self.test_scenarios = TestScenarioDesc({1: self.test_scenario1, 2: self.test_scenario2})
        self.TestAction = PicklableMock()
        self.TestAction.description = "test_function"
        self.TestAction.run = lambda x, y: x.name + y.name
        self.app_name = "test"

    def test_log_const(self):
        self.assertTrue(dialogue_manager.log_const.KEY_NAME == "key_name")
        self.assertTrue(dialogue_manager.log_const.STARTUP_VALUE == "app_startup")
        self.assertTrue(hasattr(dialogue_manager.log_const, "CHOSEN_SCENARIO_VALUE"))
        self.assertTrue(hasattr(dialogue_manager.log_const, "LAST_SCENARIO_MESSAGE"))
        self.assertTrue(not hasattr(dialogue_manager.log_const, "SCENARIO_DESCRIPTION_VAL"))

    def test_dialogue_manager_init(self):
        obj1 = dialogue_manager.DialogueManager({'scenarios': self.test_scenarios, 'external_actions': '333'},
                                                self.app_name)

        self.assertTrue(obj1.scenario_descriptions == {'scenarios': self.test_scenarios, 'external_actions': '333'})
        self.assertTrue(obj1.scenarios == self.test_scenarios)
        self.assertTrue(obj1.scenario_keys == {1, 2})
        self.assertTrue(obj1.actions == '333')
        self.assertTrue(obj1.NOTHING_FOUND_ACTION == "nothing_found_action")

    def test_dialogue_manager_found_action(self):
        obj1 = dialogue_manager.DialogueManager({'scenarios': self.test_scenarios,
                                                'external_actions': {'nothing_found_action': self.TestAction}},
                                                self.app_name)
        obj2 = dialogue_manager.DialogueManager({'scenarios': self.test_scenarios,
                                                 'external_actions': {}}, self.app_name)
        self.assertTrue(obj1._nothing_found_action().descriptions, self.TestAction.descriptions)
        with self.assertRaises(TypeError):
            obj2._nothing_found_action()

    def test_dialogue_manager_run(self):
        obj1 = dialogue_manager.DialogueManager({'scenarios': self.test_scenarios,
                                                'external_actions': {'nothing_found_action': self.TestAction}},
                                                self.app_name)
        obj2 = dialogue_manager.DialogueManager({'scenarios': self.test_scenarios,
                                                 'external_actions': {}}, self.app_name)

        # путь по умолчанию без выполнения условий
        self.assertTrue(obj1.run(self.test_text_preprocessing_result, self.test_user1) == ("TestNameResult", True))
        self.assertTrue(obj2.run(self.test_text_preprocessing_result, self.test_user1) == ("TestNameResult", True))

        # случай когда срабатоли оба условия
        self.assertTrue(obj1.run(self.test_text_preprocessing_result, self.test_user2) == ("TestNameResult", True))
        # случай, когда 2-е условие не выполнено
        self.assertTrue(obj2.run(self.test_text_preprocessing_result, self.test_user3) == ('TestNameResult', True))

        # проверка на вызов before_action, если такой задан в external_actions
        obj2.run(self.test_text_preprocessing_result, self.test_user4)
        self.assertTrue(self.test_user4.descriptions['external_actions']['before_action'].run.called)

    def test_dialogue_manager_run_scenario(self):
        obj = dialogue_manager.DialogueManager({'scenarios': self.test_scenarios,
                                                'external_actions': {'nothing_found_action': self.TestAction}},
                                               self.app_name)
        self.assertTrue(obj.run_scenario(1, self.test_text_preprocessing_result, self.test_user1) == "ResultTestName")
        self.assertTrue(obj.run_scenario(2, self.test_text_preprocessing_result, self.test_user1) == "TestNameResult")




import concurrent.futures
import json
import os
import threading
from time import sleep

from confluent_kafka.cimpl import KafkaException

import scenarios.logging.logger_constants as log_const
from core.logging.logger_utils import log
from core.message.from_message import SmartAppFromMessage
from smart_kit.start_points.main_loop_kafka import MainLoop as KafkaMainLoop


class ParallelKafkaMainLoop(KafkaMainLoop):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_workers = self.settings["template_settings"].get("max_workers", (os.cpu_count() or 1) * 5)
        self.no_workers_sleep_time = self.settings["template_settings"].get("no_workers_sleep_time", 0.05)
        self.pool = self.get_pool()
        self._tasks = []
        self._locks = {}
        log(
            "%(class_name)s started with %(workers)s workers.",
            params={
                log_const.KEY_NAME: log_const.STARTUP_VALUE, "class_name": self.__class__.__name__,
                "workers": self.max_workers,
            }
        )

    def get_pool(self) -> concurrent.futures.Executor:
        return concurrent.futures.ThreadPoolExecutor(max_workers=self.max_workers)

    def pre_handle(self):
        super().pre_handle()

        for task in list(self._tasks):
            if task.done():
                self._tasks.remove(task)

    def run(self):
        super().run()
        self.pool.shutdown()

    def process_message(self, mq_message, consumer, kafka_key, stats):
        # ну тут чутка копипасты
        mutex = None
        try:
            message_value = json.loads(mq_message.value())
            message = SmartAppFromMessage(message_value,
                                          headers=mq_message.headers(),
                                          masking_fields=self.masking_fields)
            if message.validate():
                mutex = self._locks.setdefault(message.db_uid, threading.Lock())
                mutex.acquire()

            super().process_message(mq_message, consumer, kafka_key, stats)
        except Exception:
            if mutex and mutex.locked():
                mutex.release()
            raise
        if mutex and mutex.locked():
            mutex.release()

    def iterate(self, kafka_key):
        consumer = self.consumers[kafka_key]

        mq_message = None
        message_value = None

        try:
            available_workers_count = self.max_workers - len(self._tasks)
            if available_workers_count:
                for mq_message in consumer.consume(self.max_workers - len(self._tasks)):
                    if mq_message:
                        message_value = mq_message.value()
                        task = self.pool.submit(self.process_message, mq_message, consumer, kafka_key, "")
                        self._tasks.append(task)
            else:
                sleep(self.no_workers_sleep_time)

        except KafkaException as kafka_exp:
            log("kafka error: %(kafka_exp)s. MESSAGE: {}.".format(message_value),
                params={log_const.KEY_NAME: log_const.STARTUP_VALUE,
                        "kafka_exp": str(kafka_exp),
                        log_const.REQUEST_VALUE: str(message_value)},
                level="ERROR", exc_info=True)
        except Exception:
            try:
                log("%(class_name)s iterate error. Kafka key %(kafka_key)s MESSAGE: {}.".format(message_value),
                    params={log_const.KEY_NAME: log_const.STARTUP_VALUE,
                            "kafka_key": kafka_key},
                    level="ERROR", exc_info=True)
                consumer.commit_offset(mq_message)
            except Exception:
                log("Error handling worker fail exception.",
                    level="ERROR", exc_info=True)



# coding: utf-8
import time

from scenarios.user.last_fields.last_field import LastField


class LastFields:
    def __init__(self, items, user):
        self._raw_items = items or {}
        self._items = dict()
        self._factory = LastField

    def __getitem__(self, id):
        existed_item = self._items.get(id)
        if existed_item is None:
            raw_data = self._raw_items.get(id)
            existed_item = self._factory(raw_data)
            self._items[id] = existed_item
        return existed_item

    def __setitem__(self, id, value):
        self._items[id] = value

    def __iter__(self):
        return iter(self.raw)

    def expire(self):
        items = list(self.raw.keys())
        for key in items:
            item = self.raw[key]
            remove_time = item.get("remove_time")
            if remove_time and remove_time <= int(time.time()):
                self.remove(key)

    def remove(self, key):
        if key in self._raw_items:
            self._raw_items.pop(key)
        if key in self._items:
            del self._items[key]

    def clear_all(self):
        self._raw_items.clear()
        self._items.clear()

    @property
    def raw(self):
        for i in self._items:
            item = self._items[i]
            value = getattr(item, "raw", item)
            if value:
                self._raw_items[i] = value
            else:
                if i in self._raw_items:
                    self._raw_items.pop(i)
        return self._raw_items



import time
from unittest import TestCase

from core.utils.delay_runner import DelayRunner
from smart_kit.utils.picklable_mock import PicklableMock


class TestDelayRunner(TestCase):
    def setUp(self):
        self.max_delay = 1
        self.max_delay_s = self.max_delay * 60
        self.delay_runner = DelayRunner(1)
        self.run = PicklableMock()
        self.arg = PicklableMock()
        self.delay_runner.schedule_run(self.run, [self.arg])

    def test_set_run_time(self):
        self.assertEqual(self.delay_runner._run_item, self.run)
        self.assertListEqual(self.delay_runner._run_args, [self.arg])
        self.assertTrue(0 < self.delay_runner._ts - time.time() < self.max_delay_s)

    def test_check_can_run(self):
        self.delay_runner._ts = self.delay_runner._ts - self.max_delay_s
        self.assertTrue(self.delay_runner.check_can_run())

    def test_check_cant_run(self):
        self.delay_runner._ts = time.time() + self.max_delay_s
        self.assertFalse(self.delay_runner.check_can_run())

    def test_run(self):
        self.delay_runner._ts = self.delay_runner._ts - self.max_delay_s
        self.delay_runner.run()
        self.run.assert_called_once_with(self.arg)
        self.assertLess(self.delay_runner._ts, time.time())
        self.assertIsNone(self.delay_runner._run_args)
        self.assertIsNone(self.delay_runner._run_item)



import random
from typing import Dict, Any

from core.model.base_user import BaseUser as User
from core.unified_template.unified_template import UnifiedTemplate


class ReplySelector:
    __gender_male = ["1", "male", "м"]

    def get_user_gender(self) -> str:
        if self._user_gender.lower() in self.__gender_male:
            return "male"
        else:
            return "female"

    def __init__(self, items: Dict[str, Any], user: User):
        self._bundles: dict = user.descriptions["bundles"]
        self._user_gender: str = str(user.message.payload.get("client_profile", {}).get("gender") or "")
        self.user_gender: str = self.get_user_gender()
        self.character_gender: str = str(user.message.payload.get("character", {}).get("gender") or "")
        self.character_key: str = str(user.message.payload.get("character", {}).get("name") or "")
        self.__suffix = [
            f".{self.character_key}.{self.character_gender}_to_{self.user_gender}",
            f".{self.character_gender}_to_{self.user_gender}",
            f".{self.character_key}.{self.user_gender}",
            f".{self.character_key}",
            ""
        ]
        self._user = user

    def get_text_by_key(self, bundle_name: str, reply_key="") -> str:
        result = ""
        bundle = self._bundles[bundle_name]
        if bundle:
            reply_list = None
            for suffix in self.__suffix:
                target_key = f"{reply_key}{suffix}"
                reply_list = bundle.get(target_key)
                if reply_list:
                    break
            if reply_list:
                params = self._user.parametrizer.collect()  # TODO give run-time action parameters to collect call
                result_str = random.choice(reply_list)
                result = UnifiedTemplate(result_str).render(params)
            else:
                raise KeyError("Key not found")
        return result

    @property
    def raw(self) -> None:
        return None



import inspect
import logging
import hashlib
from datetime import datetime, timezone
from functools import cached_property
from random import random
from typing import List, Optional, Dict, Any

from croniter import croniter

import core.logging.logger_constants as log_const
from core.basic_models.classifiers.basic_classifiers import Classifier, ExternalClassifier
from core.basic_models.operators.operators import Operator
from core.logging.logger_utils import log, log_classifier_result
from core.model.base_user import BaseUser
from core.model.factory import build_factory, list_factory, factory
from core.model.registered import Registered
from core.text_preprocessing.base import BaseTextPreprocessingResult
from core.text_preprocessing.preprocessing_result import TextPreprocessingResult
from core.unified_template.unified_template import UnifiedTemplate
from core.utils.stats_timer import StatsTimer
from scenarios.scenario_models.field.field_filler_description import IntersectionFieldFiller
from scenarios.user.user_model import User

requirements = Registered()

requirement_factory = build_factory(requirements)



class Requirement:
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        items = items or {}
        self._descr = items.get("_description")
        self.items = items
        self.version = items.get("version", -1)
        self.id = id
        self.is_logging_debug_mode = logging.getLogger(globals().get("__name__")).isEnabledFor(logging.getLevelName("DEBUG"))

    def _log_params(self):
        return {
            log_const.KEY_NAME: log_const.REQUIREMENT_CHECK_VALUE,
            "requirement": self.__class__.__name__
        }

    def _check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return True

    def _on_check_error_result(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser) -> bool:
        # return True or False if it's acceptable in custom requirement
        raise

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        try:
            result = self._check(text_preprocessing_result, user, params)
        except:
            return self.on_check_error(text_preprocessing_result, user)
        if self.is_logging_debug_mode:
            log_params = self._log_params()
            log_params[log_const.KEY_NAME] = log_const.REQUIREMENT_TRACE_VALUE
            log_params["result"] = result
            log_params["raw_items"] = str(self.items)
            log_params["params"] = params
            if self._descr:
                log_params["descr"] = self._descr
            log(f"TRACING %(requirement)s with id {self.id}: Result is %(result)s.", user, log_params, level="DEBUG")
        return result

    def on_check_error(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser):
        result = self._on_check_error_result(text_preprocessing_result, user)
        log_params = self._log_params()
        log_params["masked_message"]: str(user.message.masked_value)
        log_params["result"]: result
        log("exc_handler: Requirement FAILED to check. Return %(result)s. MESSAGE: %(masked_message)s.",
            user, log_params,
            level="ERROR", exc_info=True)
        return result


class CompositeRequirement(Requirement):
    requirements: List[Requirement]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(CompositeRequirement, self).__init__(items, id)
        self._requirements = items["requirements"]
        self.requirements = self.build_requirements()

    @list_factory(Requirement)
    def build_requirements(self):
        return self._requirements


class AndRequirement(CompositeRequirement):

    def _check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return all(requirement.check(text_preprocessing_result=text_preprocessing_result, user=user, params=params)
                   for requirement in self.requirements)


class OrRequirement(CompositeRequirement):

    def _check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return any(requirement.check(text_preprocessing_result=text_preprocessing_result, user=user, params=params)
                   for requirement in self.requirements)


class NotRequirement(Requirement):
    requirement: Requirement

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(NotRequirement, self).__init__(items, id)
        self._requirement = items["requirement"]
        self.requirement = self.build_requirement()

    @factory(Requirement)
    def build_requirement(self):
        return self._requirement

    def _check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return not self.requirement.check(text_preprocessing_result=text_preprocessing_result, user=user, params=params)


class ComparisonRequirement(Requirement):
    operator: Operator

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(ComparisonRequirement, self).__init__(items, id)
        self._operator = items["operator"]
        self.operator = self.build_operator()

    @factory(Operator)
    def build_operator(self):
        return self._operator


class RandomRequirement(Requirement):
    percent: int

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(RandomRequirement, self).__init__(items, id)
        self.percent = items["percent"]

    def _check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        result = random() * 100
        return result < self.percent


class TopicRequirement(Requirement):
    topics: List[str]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(TopicRequirement, self).__init__(items, id)
        self.topics = items["topics"]

    def _check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return user.message.topic_key in self.topics


class TemplateRequirement(Requirement):
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(TemplateRequirement, self).__init__(items, id)
        self._template = UnifiedTemplate(items["template"])

    def _check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        params = params or {}
        collected = user.parametrizer.collect(text_preprocessing_result)
        params.update(collected)
        render_result = self._template.render(params)
        if render_result == "True":
            return True
        if render_result == "False":
            return False
        raise TypeError(f'Template result should be "True" or "False", got: '
                        f'{render_result} for template {self.items["template"]}')


class RollingRequirement(Requirement):
    percent: int

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(RollingRequirement, self).__init__(items, id)
        self.percent = items["percent"]

    def _check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        id = user.id
        s = id.encode('utf-8')
        hash = int(hashlib.sha256(s).hexdigest(), 16)
        res = hash % 100
        return res < self.percent


class TimeRequirement(ComparisonRequirement):
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super().__init__(items, id)

    def _check(
            self,
            text_preprocessing_result: BaseTextPreprocessingResult,
            user: BaseUser,
            params: Dict[str, Any] = None
    ) -> bool:
        message_time_dict = user.message.payload['meta']['time']
        message_timestamp_sec = message_time_dict['timestamp'] // 1000
        message_time = datetime.fromtimestamp(message_timestamp_sec, tz=timezone.utc).time()
        return self.operator.compare(message_time)

    @factory(Operator)
    def build_operator(self):
        operator = dict(self._operator)
        amount_time = datetime.strptime(operator["amount"], '%H:%M:%S').time()
        operator["amount"] = amount_time
        return operator


class DateTimeRequirement(Requirement):
    match_cron: str

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super().__init__(items, id)
        self.match_cron = items['match_cron']

    def _check(
            self,
            text_preprocessing_result: BaseTextPreprocessingResult,
            user: BaseUser,
            params: Dict[str, Any] = None
    ) -> bool:
        message_time_dict = user.message.payload['meta']['time']
        message_timestamp_sec = message_time_dict['timestamp'] // 1000
        message_datetime = datetime.fromtimestamp(message_timestamp_sec)
        return croniter.match(self.match_cron, message_datetime)


class IntersectionRequirement(Requirement):
    phrases: Optional[List]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super().__init__(items, id)
        self.filler = IntersectionFieldFiller(
            {
                'cases': {
                    True: items.get('phrases', []),
                },
                'default': False,
            },
            id,
        )

    def _check(
            self,
            text_preprocessing_result: TextPreprocessingResult,
            user: User,
            params: Dict[str, Any] = None
    ) -> bool:
        result = bool(
            self.filler.extract(text_preprocessing_result, user, params),
        )
        return result


class ClassifierRequirement(Requirement):
    """Условие, которое зависит от результата классификации.
    Возвращает True, если результат классификации запроса относится к одной из указанных категорий, прошедших порог,
    но не равной классу other.
    """

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(ClassifierRequirement, self).__init__(items=items, id=id)
        self._classifier = items["classifier"]

    @cached_property
    def classifier(self) -> Classifier:
        return ExternalClassifier(self._classifier)

    def _check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        check_res = True
        classifier = self.classifier
        with StatsTimer() as timer:
            classification_res = classifier.find_best_answer(
                text_preprocessing_result, scenario_classifiers=user.descriptions["external_classifiers"])

        log_classifier_result(classification_res, user, classifier, timer)

        if not classification_res or classification_res[0][classifier.class_other]:
            check_res = False

        return check_res


class FormFieldValueRequirement(Requirement):
    """Условие возвращает True, если в форме form_name в поле field_name значение совпадает с переданным value,
    иначе - False. Данное условие предназначено только для плоских форм.
    """

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(FormFieldValueRequirement, self).__init__(items, id)
        self.form_name = items["form_name"]
        self.field_name = items["field_name"]
        self.value = items["value"]

    def _check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
              params: Dict[str, Any] = None) -> bool:
        return user.forms[self.form_name].fields[self.field_name].value == self.value


class EnvironmentRequirement(Requirement):
    """Условие возвращает True, если сценарий исполняется на стенде из числа values, иначе - False.
    Так, например, можно ограничить сценарий для исполнения только на тестовых средах.
    Возможные значения в values: ift, uat, pt, prod (это ИФТ, ПСИ, НТ, ПРОМ).
    """

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(EnvironmentRequirement, self).__init__(items, id)
        self.values = items.get("values", [])
        # Из конфига получаем среду исполнения
        from smart_kit.configs import get_app_config
        app_config = get_app_config()
        self.environment = app_config.ENVIRONMENT
        # Если среда исполнения задана, то проверям, что среда в списке возможных значений для сценария, иначе - False
        self.check_result = self.environment in self.values if self.environment else False

    def _check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return self.check_result


class CharacterIdRequirement(Requirement):
    """Условие возвращает True, если идентификатор выбранного персонажа входит
    в список значений, иначе - False.
    """

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(CharacterIdRequirement, self).__init__(items=items, id=id)
        self.values = items["values"]

    def _check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
              params: Dict[str, Any] = None) -> bool:
        return user.message.payload["character"]["id"] in self.values


class FeatureToggleRequirement(Requirement):
    """Условие возвращает True, если проверка указанного тогла по названию возвращает True, иначе - False.
    Тоглы задаются в template_config.yml, с помощью значений True и False их можно включить или выключить.
    """

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(FeatureToggleRequirement, self).__init__(items=items, id=id)
        self.toggle_name = items["toggle_name"]

    def _check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
              params: Dict[str, Any] = None) -> bool:
        return user.settings["template_settings"].get(self.toggle_name, False)



# -*- coding: utf-8 -*-
import sys
import unittest

from core.model.field import Field
from core.model.model import Model


class MockField():
    def __init__(self, value, description, *args):
        self.value = value
        self.descr = description

    @property
    def raw(self):
        return self.value

    def fill(self, origin_value):
        self.value = origin_value
        return True


class MockModel(Model):
    @property
    def fields(self):
        return [Field("field1", MockField, "field1_descr")]


class ModelTest(unittest.TestCase):
    def setUp(self):
        self.model = MockModel(values={"field1": "field1_data"}, user=None)

    def test_get_field(self):
        field1 = self.model.get_field("field1")
        assert field1.value, field1.descr == ("field1_data", "field1_descr")

    def test_get_field_error(self):
        exception = None
        try:
            self.model.get_field("field_unknown")
        except AttributeError:
            exception = sys.exc_info()[0]
        assert exception == AttributeError

    def test_raw(self):
        raw = self.model.raw
        assert raw == dict(field1="field1_data")



import json

import core.basic_models.operators.comparators as cmp
import core.basic_models.operators.operators as op
import core.basic_models.requirement.device_requirements as dr
import scenarios.scenario_models.field.composite_fillers as cffd
import scenarios.scenario_models.field.field_filler_description as ffd
import scenarios.scenario_models.field_requirements.field_requirements as frd
from core.basic_models.actions.basic_actions import actions, action_factory, Action, \
    DoingNothingAction, RequirementAction, ChoiceAction, ElseAction, CompositeAction
from core.basic_models.actions.client_profile_actions import GiveMeMemoryAction, RememberThisAction
from core.basic_models.actions.counter_actions import CounterIncrementAction, CounterDecrementAction, \
    CounterClearAction, CounterSetAction, CounterCopyAction
from core.basic_models.actions.external_actions import ExternalAction
from core.basic_models.actions.external_actions import ExternalActions
from core.basic_models.actions.push_action import PushAction, PUSH_NOTIFY
from core.basic_models.actions.smartpay import SmartPayCreateAction, SmartPayPerformAction, SmartPayGetStatusAction, \
    SmartPayConfirmAction, SmartPayDeleteAction, SmartPayRefundAction
from core.basic_models.actions.string_actions import StringAction, AfinaAnswerAction, SDKAnswer, \
    SDKAnswerToUser
from core.basic_models.actions.variable_actions import ClearVariablesAction, DeleteVariableAction, \
    SetLocalVariableAction, SetVariableAction
from core.basic_models.answer_items.answer_items import items_factory, SdkAnswerItem, answer_items, BubbleText, \
    ItemCard, PronounceText, SuggestText, SuggestDeepLink, RawItem
from core.basic_models.classifiers.basic_classifiers import classifiers, classifier_factory, Classifier, \
    ExternalClassifier, SciKitClassifier, SkipClassifier
from core.basic_models.classifiers.external_classifiers import ExternalClassifiers
from core.basic_models.requirement.basic_requirements import requirement_factory, IntersectionRequirement
from core.basic_models.requirement.basic_requirements import requirements, Requirement, AndRequirement, \
    OrRequirement, NotRequirement, TemplateRequirement, RandomRequirement, TimeRequirement, DateTimeRequirement, \
    ClassifierRequirement, FormFieldValueRequirement, EnvironmentRequirement, CharacterIdRequirement, \
    FeatureToggleRequirement
from core.basic_models.requirement.counter_requirements import CounterValueRequirement, CounterUpdateTimeRequirement
from core.basic_models.requirement.device_requirements import ChannelRequirement
from core.basic_models.requirement.external_requirements import ExternalRequirement
from core.basic_models.requirement.external_requirements import ExternalRequirements
from core.basic_models.requirement.project_requirements import SettingsRequirement
from core.basic_models.requirement.user_text_requirements import AnySubstringInLoweredTextRequirement, \
    IntersectionWithTokensSetRequirement, NormalizedTextInSetRequirement, \
    PhoneNumberNumberRequirement, NumInRangeRequirement
from core.basic_models.scenarios.base_scenario import BaseScenario
from core.basic_models.scenarios.base_scenario import scenarios
from core.configs.base_config import BaseConfig
from core.db_adapter.aioredis_adapter import AIORedisAdapter
from core.db_adapter.db_adapter import db_adapters
from core.db_adapter.ignite_adapter import IgniteAdapter
from core.db_adapter.ignite_thread_adapter import IgniteThreadAdapter
from core.db_adapter.memory_adapter import MemoryAdapter
from core.db_adapter.redis_adapter import RedisAdapter
from core.descriptions.descriptions import registered_description_factories
from core.model.queued_objects.limited_queued_hashable_objects_description import \
    LimitedQueuedHashableObjectsDescriptionsItems
from core.model.registered import registered_factories
from core.repositories.classifier_repository import ClassifierRepository
from core.repositories.file_repository import FileRepository
from core.repositories.folder_repository import FolderRepository
from core.request.base_request import requests_registered
from core.request.rest_request import RestRequest
from core.utils.loader import ordered_json
from scenarios.actions.action import (
    BreakScenarioAction, ChoiceScenarioAction, ClearCurrentScenarioAction, ClearCurrentScenarioFormAction,
    ClearFormAction, ClearInnerFormAction, ClearScenarioByIdAction, CompositeFillFieldAction, FillFieldAction,
    RemoveCompositeFormFieldAction, RemoveFormFieldAction, SaveBehaviorAction, ResetCurrentNodeAction,
    RunScenarioAction, RunLastScenarioAction, AddHistoryEventAction, ClearAllScenariosAction
)
from scenarios.actions.action import ProcessBehaviorAction, SelfServiceActionWithState, EmptyAction
from scenarios.behaviors.behavior_descriptions import BehaviorDescriptions
from scenarios.requirements.requirements import AskAgainExistRequirement, TemplateInArrayRequirement, \
    ArrayItemInTemplateRequirement, RegexpInTemplateRequirement
from scenarios.scenario_descriptions.form_filling_scenario import FormFillingScenario
from scenarios.scenario_descriptions.scenarios_description import ScenariosDescriptions
from scenarios.scenario_descriptions.tree_scenario.tree_scenario import TreeScenario
from scenarios.scenario_models.field.composite_field import CompositeField
from scenarios.scenario_models.field.external_field_filler_descriptions import ExternalFieldFillerDescriptions
from scenarios.scenario_models.field.field import field_models, field_model_factory, BasicField, IntegrationField, \
    QuestionField
from scenarios.scenario_models.field.field_descriptions.basic_field_description import field_descriptions, \
    field_description_factory, BasicFieldDescription
from scenarios.scenario_models.field.field_descriptions.composite_field_description import CompositeFieldDescription
from scenarios.scenario_models.field.field_descriptions.integration_field_description import IntegrationFieldDescription
from scenarios.scenario_models.field.field_descriptions.question_field_description import QuestionFieldDescription
from scenarios.scenario_models.forms.composite_forms import CompositeForm
from scenarios.scenario_models.forms.form import BaseForm, form_models, form_model_factory, Form
from scenarios.scenario_models.forms.form_description import form_descriptions, form_description_factory, \
    BaseFormDescription, FormDescription, CompositeFormDescription
from scenarios.scenario_models.forms.forms_description import FormsDescription
from scenarios.scenario_models.history import HistoryDescription, \
    EventFormatter, formatters, formatters_factory, HistoryEventFormatter
from scenarios.scenario_models.scenario_models import scenario_models, scenario_model_factory, BaseScenarioModel, \
    TreeScenarioModel, FormFillingScenarioModel
from scenarios.user.last_scenarios.last_scenarios_descriptions import LastScenariosDescriptionsItems
from scenarios.user.preprocessing_messages.preprocessing_messages_description import \
    PreprocessingMessagesDescription
from smart_kit.action.http import HTTPRequestAction
from smart_kit.action.smart_geo_action import SmartGeoAction
from smart_kit.message.get_to_message import to_messages
from smart_kit.message.as_is_to_message import AsIsToMessage
from smart_kit.request.kafka_request import SmartKitKafkaRequest

from core.db_adapter.aioredis_sentinel_adapter import AIORedisSentinelAdapter


class SmartAppResources(BaseConfig):
    def __init__(self, source, references_path, settings):
        super(SmartAppResources, self).__init__(source=source)
        self.references_path = references_path
        self.repositories = [
            FolderRepository(self.subfolder_path("forms"), loader=ordered_json, source=source,
                             key="forms"),
            FolderRepository(self.subfolder_path("scenarios"), loader=ordered_json, source=source,
                             key="scenarios"),
            FileRepository(self.subfolder_path("preprocessing_messages_for_scenarios_settings.json"),
                           loader=ordered_json,
                           source=source, key="preprocessing_messages_for_scenarios"),
            FileRepository(self.subfolder_path("last_scenarios_descriptions.json"), loader=ordered_json,
                           source=source, key="last_scenarios"),
            FileRepository(self.subfolder_path("history.json"), loader=ordered_json, source=source,
                           key="history"),
            FolderRepository(self.subfolder_path("behaviors"), loader=ordered_json, source=source,
                             key="behaviors"),
            FolderRepository(self.subfolder_path("actions"), loader=ordered_json, source=source,
                             key="external_actions"),
            FolderRepository(self.subfolder_path("requirements"), loader=ordered_json, source=source,
                             key="external_requirements"),
            FolderRepository(self.subfolder_path("field_fillers"), loader=ordered_json, source=source,
                             key="external_field_fillers"),
            FileRepository(self.subfolder_path("responses.json"), loader=ordered_json, source=source,
                           key="responses"),
            FileRepository(self.subfolder_path("last_action_ids.json"), loader=ordered_json,
                           source=source, key="last_action_ids"),
            FolderRepository(self.subfolder_path("bundles"), loader=ordered_json, source=source,
                             key="bundles"),
            ClassifierRepository(
                description_path=self.subfolder_path("classifiers"),
                data_path=self.subfolder_path("classifiers_data"),
                loader=json.loads,
                source=source,
                key="external_classifiers"
            )
        ]

        self.repositories = self.override_repositories(self.repositories)
        self.init()

    @property
    def _subfolder(self):
        return self.references_path

    def override_repositories(self, repositories: list):
        """
        Метод предназначен для переопределения репозиториев в дочерних классах.
        :param repositories: Список репозиториев родителя
        :return: Переопределённый в наследниках список репозиториев
        """
        return repositories

    def init(self):
        super(SmartAppResources, self).init()
        self.init_factories()
        self.init_field_filler_description()
        self.init_scenarios()
        self.init_field_requirements()
        self.init_repo_factories()
        self.init_scenario_models()
        self.init_form_descriptions()
        self.init_form_model()
        self.init_field_descriptions()
        self.init_field_model()
        self.init_actions()
        self.init_requirements()
        self.init_operators()
        self.init_comparators()
        self.init_sdk_items()
        self.init_history_formatters()
        self.init_db_adapters()
        self.init_classifiers()
        self.init_message_handlers()

    def init_field_requirements(self):
        frd.field_requirements[None] = frd.FieldRequirement
        frd.field_requirements["and"] = frd.AndFieldRequirement
        frd.field_requirements["comparison"] = frd.ComparisonFieldRequirement
        frd.field_requirements["field_text_length"] = frd.TextLengthFieldRequirement
        frd.field_requirements["is_int"] = frd.IsIntFieldRequirement
        frd.field_requirements["not"] = frd.NotFieldRequirement
        frd.field_requirements["or"] = frd.OrFieldRequirement
        frd.field_requirements["token_part_in_set"] = frd.TokenPartInSet
        frd.field_requirements["value_in_set"] = frd.ValueInSetRequirement

    def init_field_filler_description(self):
        ffd.field_filler_description[None] = ffd.FieldFillerDescription
        ffd.field_filler_description["all_regexps"] = ffd.AllRegexpsFieldFiller
        ffd.field_filler_description["approve"] = ffd.ApproveFiller
        ffd.field_filler_description["approve_strictly"] = ffd.ApproveRawTextFiller
        ffd.field_filler_description["available_info_filler"] = ffd.AvailableInfoFiller
        ffd.field_filler_description["choice"] = cffd.ChoiceFiller
        ffd.field_filler_description["classifier"] = ffd.ClassifierFiller
        ffd.field_filler_description["classifier_meta"] = ffd.ClassifierFillerMeta
        ffd.field_filler_description["composite"] = ffd.CompositeFiller
        ffd.field_filler_description["currency_first"] = ffd.FirstCurrencyFiller
        ffd.field_filler_description["else"] = cffd.ElseFiller
        ffd.field_filler_description["external"] = ffd.ExternalFieldFillerDescription
        ffd.field_filler_description["geo"] = ffd.FirstGeoFiller
        ffd.field_filler_description["get_first_person"] = ffd.FirstPersonFiller
        ffd.field_filler_description["intersection"] = ffd.IntersectionFieldFiller
        ffd.field_filler_description["intersection_original_text"] = ffd.IntersectionOriginalTextFiller
        ffd.field_filler_description["number_first"] = ffd.FirstNumberFiller
        ffd.field_filler_description["organisation"] = ffd.FirstOrgFiller
        ffd.field_filler_description["previous_messages_filler"] = ffd.PreviousMessagesFiller
        ffd.field_filler_description["regexp"] = ffd.RegexpFieldFiller
        ffd.field_filler_description["regexp_string_operations"] = ffd.RegexpAndStringOperationsFieldFiller
        ffd.field_filler_description["requirement"] = cffd.RequirementFiller
        ffd.field_filler_description["user_id"] = ffd.UserIdFiller
        ffd.field_filler_description["date_period_filler"] = ffd.DatePeriodFiller

    def init_scenarios(self):
        scenarios[None] = BaseScenario
        scenarios["base"] = BaseScenario
        scenarios["form_filling"] = FormFillingScenario
        scenarios["tree"] = TreeScenario

    def init_form_descriptions(self):
        form_descriptions[None] = FormDescription
        form_descriptions["base"] = FormDescription
        form_descriptions["composite"] = CompositeFormDescription

    def init_form_model(self):
        form_models[FormDescription] = Form
        form_models[CompositeFormDescription] = CompositeForm

    def init_field_descriptions(self):
        field_descriptions[None] = QuestionFieldDescription
        field_descriptions["question"] = QuestionFieldDescription
        field_descriptions["composite"] = CompositeFieldDescription
        field_descriptions["integration"] = IntegrationFieldDescription

    def init_field_model(self):
        field_models[BasicFieldDescription] = BasicField
        field_models[CompositeFieldDescription] = CompositeField
        field_models[QuestionFieldDescription] = QuestionField
        field_models[IntegrationFieldDescription] = IntegrationField

    def init_scenario_models(self):
        scenario_models[FormFillingScenario] = FormFillingScenarioModel
        scenario_models[TreeScenario] = TreeScenarioModel

    def init_factories(self):
        registered_factories[Action] = action_factory
        registered_factories[ffd.FieldFillerDescription] = ffd.field_filler_factory
        registered_factories[frd.FieldRequirement] = frd.field_requirement_factory
        registered_factories[BaseScenarioModel] = scenario_model_factory
        registered_factories[BaseFormDescription] = form_description_factory
        registered_factories[BaseForm] = form_model_factory
        registered_factories[BasicFieldDescription] = field_description_factory
        registered_factories[BasicField] = field_model_factory
        registered_factories[op.Operator] = op.operator_factory
        registered_factories[cmp.Comparator] = cmp.comparator_factory
        registered_factories[Requirement] = requirement_factory
        registered_factories[SdkAnswerItem] = items_factory
        registered_factories[EventFormatter] = formatters_factory
        registered_factories[Classifier] = classifier_factory

    def init_repo_factories(self):
        registered_description_factories["forms"] = FormsDescription
        registered_description_factories["scenarios"] = ScenariosDescriptions
        registered_description_factories["preprocessing_messages_for_scenarios"] = PreprocessingMessagesDescription
        registered_description_factories["last_scenarios"] = LastScenariosDescriptionsItems
        registered_description_factories["external_actions"] = ExternalActions
        registered_description_factories["behaviors"] = BehaviorDescriptions
        registered_description_factories["history"] = HistoryDescription
        registered_description_factories["external_requirements"] = ExternalRequirements
        registered_description_factories["external_field_fillers"] = ExternalFieldFillerDescriptions
        registered_description_factories["last_action_ids"] = LimitedQueuedHashableObjectsDescriptionsItems
        registered_description_factories["external_classifiers"] = ExternalClassifiers

    def init_actions(self):
        actions[None] = EmptyAction
        actions["add_history_event"] = AddHistoryEventAction
        actions["break_scenario"] = BreakScenarioAction
        actions["choice"] = ChoiceAction
        actions["choice_scenario"] = ChoiceScenarioAction
        actions["clear_current_scenario"] = ClearCurrentScenarioAction
        actions["clear_all_scenarios"] = ClearAllScenariosAction
        actions["clear_current_scenario_form"] = ClearCurrentScenarioFormAction
        actions["clear_form_by_id"] = ClearFormAction
        actions["clear_inner_form_by_id"] = ClearInnerFormAction
        actions["clear_scenario_by_id"] = ClearScenarioByIdAction
        actions["clear_variables"] = ClearVariablesAction
        actions["composite"] = CompositeAction
        actions["composite_fill_field"] = CompositeFillFieldAction
        actions["counter_clear"] = CounterClearAction
        actions["counter_copy"] = CounterCopyAction
        actions["counter_decrement"] = CounterDecrementAction
        actions["counter_increment"] = CounterIncrementAction
        actions["counter_set"] = CounterSetAction
        actions["delete_variable"] = DeleteVariableAction
        actions["do_nothing"] = DoingNothingAction
        actions["else"] = ElseAction
        actions["external"] = ExternalAction
        actions["fill_field"] = FillFieldAction
        actions["http_request"] = HTTPRequestAction
        actions["process_behavior"] = ProcessBehaviorAction
        actions["random_field_answer"] = AfinaAnswerAction
        actions["remove_composite_form_field"] = RemoveCompositeFormFieldAction
        actions["remove_form_field"] = RemoveFormFieldAction
        actions["requirement"] = RequirementAction
        actions["reset_current_node"] = ResetCurrentNodeAction
        actions["run_last_scenario"] = RunLastScenarioAction
        actions["run_scenario"] = RunScenarioAction
        actions["save_behavior"] = SaveBehaviorAction
        actions["sdk_answer"] = SDKAnswer
        actions["sdk_answer_to_user"] = SDKAnswerToUser
        actions["self_service_with_state"] = SelfServiceActionWithState
        actions["set_local_variable"] = SetLocalVariableAction
        actions["set_variable"] = SetVariableAction
        actions["string"] = StringAction
        actions["push"] = PushAction
        actions["give_me_memory"] = GiveMeMemoryAction
        actions["remember_this"] = RememberThisAction
        actions["smart_geo"] = SmartGeoAction
        actions["smartpay_create"] = SmartPayCreateAction
        actions["smartpay_perform"] = SmartPayPerformAction
        actions["smartpay_get_status"] = SmartPayGetStatusAction
        actions["smartpay_confirm"] = SmartPayConfirmAction
        actions["smartpay_delete"] = SmartPayDeleteAction
        actions["smartpay_refund"] = SmartPayRefundAction

    def init_requirements(self):
        requirements[None] = Requirement
        requirements["and"] = AndRequirement
        requirements["any_substring_in_lowered_text"] = AnySubstringInLoweredTextRequirement
        requirements["app_type"] = dr.AppTypeRequirement
        requirements["array_item_in_template"] = ArrayItemInTemplateRequirement
        requirements["ask_again_exist"] = AskAgainExistRequirement
        requirements["capabilities_property_available"] = dr.CapabilitiesPropertyAvailableRequirement
        requirements["channel"] = ChannelRequirement
        requirements["character_id"] = CharacterIdRequirement
        requirements["classifier"] = ClassifierRequirement
        requirements["counter_time"] = CounterUpdateTimeRequirement
        requirements["counter_value"] = CounterValueRequirement
        requirements["datetime"] = DateTimeRequirement
        requirements["environment"] = EnvironmentRequirement
        requirements["external"] = ExternalRequirement
        requirements["feature_toggle_check"] = FeatureToggleRequirement
        requirements["form_field_value"] = FormFieldValueRequirement
        requirements["intersection"] = IntersectionRequirement
        requirements["intersection_with_tokens"] = IntersectionWithTokensSetRequirement
        requirements["normalized_text_in_set"] = NormalizedTextInSetRequirement
        requirements["not"] = NotRequirement
        requirements["num_in_range"] = NumInRangeRequirement
        requirements["or"] = OrRequirement
        requirements["phone_number_number"] = PhoneNumberNumberRequirement
        requirements["platform_type"] = dr.PlatformTypeRequirement
        requirements["platform_version"] = dr.PlatformVersionRequirement
        requirements["random"] = RandomRequirement
        requirements["regexp_in_template"] = RegexpInTemplateRequirement
        requirements["surface"] = dr.SurfaceRequirement
        requirements["surface_version"] = dr.SurfaceVersionRequirement
        requirements["template"] = TemplateRequirement
        requirements["template_in_array"] = TemplateInArrayRequirement
        requirements["time"] = TimeRequirement
        requirements["settings"] = SettingsRequirement

    def init_sdk_items(self):
        answer_items["bubble_text"] = BubbleText
        answer_items["item_card"] = ItemCard
        answer_items["pronounce_text"] = PronounceText
        answer_items["raw"] = RawItem
        answer_items["suggest_deeplink"] = SuggestDeepLink
        answer_items["suggest_text"] = SuggestText

    def init_operators(self):
        op.operators["any"] = op.AnyOperator
        op.operators["composite"] = op.CompositeOperator
        op.operators["endswith"] = op.EndsWithOperator
        op.operators["equal"] = op.EqualOperator
        op.operators["exists"] = op.Exists
        op.operators["in"] = op.InOperator
        op.operators["less"] = op.LessOperator
        op.operators["less_or_equal"] = op.LessOrEqualOperator
        op.operators["more"] = op.MoreOperator
        op.operators["more_or_equal"] = op.MoreOrEqualOperator
        op.operators["not_equal"] = op.NotEqualOperator
        op.operators["startswith"] = op.StartsWithOperator

    def init_comparators(self):
        cmp.comparators["equal"] = cmp.EqualComparator
        cmp.comparators["less"] = cmp.LessComparator
        cmp.comparators["less_or_equal"] = cmp.LessOrEqualComparator
        cmp.comparators["more"] = cmp.MoreComparator
        cmp.comparators["more_or_equal"] = cmp.MoreOrEqualComparator
        cmp.comparators["not_equal"] = cmp.NotEqualComparator

    def init_history_formatters(self):
        formatters[None] = HistoryEventFormatter
        formatters["history_formatter_20"] = HistoryEventFormatter

    def init_requests(self):
        requests_registered[None] = SmartKitKafkaRequest
        requests_registered["kafka"] = SmartKitKafkaRequest
        requests_registered["rest"] = RestRequest

    def init_db_adapters(self):
        db_adapters[None] = MemoryAdapter
        db_adapters["ignite"] = IgniteAdapter
        db_adapters["ignite_thread"] = IgniteThreadAdapter
        db_adapters["memory"] = MemoryAdapter
        db_adapters["redis"] = RedisAdapter
        db_adapters["aioredis"] = AIORedisAdapter
        db_adapters["aioredis_sentinel"] = AIORedisSentinelAdapter

    def init_classifiers(self):
        classifiers[None] = Classifier
        classifiers["external"] = ExternalClassifier
        classifiers["scikit"] = SciKitClassifier
        classifiers["skip"] = SkipClassifier

    def init_message_handlers(self):
        to_messages[PUSH_NOTIFY] = AsIsToMessage
